"""
Automates updating the static files.
Now saves both the raw CSV and the generated JSON files.
If new files need to be added, then place them in the TARGETS list.
"""
import aiohttp
import asyncio
import json
import logging
import csv
import os
import zipfile
import zstandard
import lzma
from pathlib import Path

class StaticUpdater:
    def __init__(self):

        self.TARGETS = []
        self.supported_languages: list[str] = []
        self.USED_TIDS = set()

        # keep the raw CSV files
        self.KEEP_CSV = False
        # keep the raw JSON files
        self.KEEP_JSON = False
        # removes any TIDs not used in the static files
        self.PRUNE_TRANSLATIONS = True
        # base path for the static files to be stored in
        self.BASE_PATH = ""

        self.FINGERPRINT = "475cb6a2d13043762034ddd6a198bad23e0782eb"
        self.CLASH_VERSION = "" or "latest"
        self.VERSION_PARAM = "version" if self.CLASH_VERSION == "latest" else "versionCode"
        self.APK_URL = f"https://d.apkpure.net/b/APK/com.supercell.clashofclans?{self.VERSION_PARAM}={self.CLASH_VERSION}"

        self.translation_data = {}
        self.full_building_data = {}
        self.full_supercharges_data = {}
        self.full_abilities_data = {}
        self.full_hero_data = {}
        self.full_townhall_data = {}
        self.full_troop_data = {}
        self.full_resource_data = {}

        self.lab_to_townhall = {}
        self.smithy_to_townhall = {}
        self.bb_lab_to_townhall = {}
        self.pethouse_to_townhall = {}

    async def download(self, url: str, as_json: bool = False):
        async with aiohttp.request('GET', url) as fp:
            if as_json:
                c = await fp.json()
            else:
                c = await fp.read()
        return c

    async def get_fingerprint(self):
        data = await self.download(self.APK_URL)

        with open("apk.zip", "wb") as f:
            f.write(data)
        zf = zipfile.ZipFile("apk.zip")
        with zf.open('assets/fingerprint.json') as fp:
            fingerprint = json.loads(fp.read())['sha']

        os.remove("apk.zip")
        return fingerprint

    def decompress(self, data):
        """
        Decompresses the given bytes 'data' if needed (LZHAM, ZSTD, or LZMA).
        Returns (decompressed_bytes, compression_details).
        """
        if data[0:4] == b"SCLZ":
            logging.debug("Decompressing using LZHAM ...")
            import lzham

            dict_size = int.from_bytes(data[4:5], byteorder="big")
            uncompressed_size = int.from_bytes(data[5:9], byteorder="little")

            decompressed = lzham.decompress(data[9:], uncompressed_size, {"dict_size_log2": dict_size})
            return decompressed, {
                "dict_size": dict_size,
                "uncompressed_size": uncompressed_size,
            }

        if int.from_bytes(data[0:4], byteorder="little") == zstandard.MAGIC_NUMBER:
            logging.debug("Decompressing using ZSTD ...")
            decompressed = zstandard.decompress(data)
            return decompressed, {
                "dict_size": None,
                "uncompressed_size": None,
            }

        # Otherwise, assume LZMA
        logging.debug("Decompressing using LZMA ...")
        data = data[0:9] + (b"\x00" * 4) + data[9:]
        prop = data[0]
        o_prop = prop
        if prop > (4 * 5 + 4) * 9 + 8:
            raise Exception("LZMA properties error")
        decompressed = lzma.LZMADecompressor().decompress(data)
        return decompressed, {"lzma_prop": o_prop}

    def process_csv(self, data, file_path, save_name):
        """
        1. Decompress data -> raw CSV
        2. Write raw CSV to disk
        3. Parse disk CSV -> final_data with levels:
        - If columns[1] is an int-type (e.g. “Level”), use that for your level keys.
        - Otherwise auto-enumerate each row under the same troop as levels “1”, “2”, …
        4. Post-process:
        - Promote any column that only appears in level “1” up to troop-level.
        - If a troop ends up with only one level, flatten it.
        5. Write out JSON with (troop → levels + troop-level props).
        6. Delete the CSV file.
        """
        # Check if data is compressed
        if self.is_compressed(data):
            # Strip signature header if present (Sig: + 64 bytes signature)
            if data[:4] == b"Sig:":
                logging.debug("Stripping Sig: header and signature...")
                data = data[68:]  # Skip "Sig:" (4 bytes) + 64-byte signature
            decompressed_data, _ = self.decompress(data)
        else:
            decompressed_data = data

        # 1) Write out the raw CSV
        with open(file_path, "wb") as f:
            f.write(decompressed_data)

        # 2) Load rows and grab the header + type row
        with open(file_path, encoding="utf-8") as csvf:
            rows = list(csv.reader(csvf))
        if len(rows) < 2:
            with open(f"{save_name}.json", "w", encoding="utf-8") as jf:
                jf.write("{}")
            os.remove(file_path)
            return

        columns   = rows[0]
        types_row = rows[1]

        # detect if col[1] really is a numeric level
        is_numeric_level = (
            types_row[1].lower() == "int"
            or "level" in columns[1].lower()
        )

        final_data     = {}
        current_troop  = None
        level_counter  = None
        current_level  = None

        # 3) Parse & build final_data
        for row in rows[2:]:
            if not any(cell.strip() for cell in row):
                continue

            # new troop?
            if row[0].strip():
                current_troop = row[0].strip()
                final_data[current_troop] = {}
                # reset counters
                if not is_numeric_level:
                    level_counter = 1
                current_level = None

            if current_troop is None:
                continue

            # pick level key
            if is_numeric_level:
                # existing behavior: use col[1]
                if len(row) > 1 and row[1].strip():
                    current_level = row[1].strip()
                if not current_level:
                    continue
                lvl_key = current_level
            else:
                # auto-enumerate each data-row
                lvl_key = str(level_counter)
                level_counter += 1

            # grab/create dict for this level
            level_dict = final_data[current_troop].setdefault(lvl_key, {})

            # fill in columns
            for idx, col_name in enumerate(columns):
                if idx >= len(row):
                    break
                val = row[idx].strip()
                if val == "":
                    continue
                low = val.lower()
                if low == "true":
                    conv = True
                elif low == "false":
                    conv = False
                elif val.isdigit() or (val.startswith("-") and val[1:].isdigit()):
                    conv = int(val)
                else:
                    conv = val
                level_dict[col_name] = conv

        # 4a) Promote any base-level-only columns up to troop-level
        for troop, levels in list(final_data.items()):
            lvl_keys = sorted(levels.keys(), key=lambda x: int(x) if x.isdigit() else 999999)
            if len(lvl_keys) <= 1:
                continue
            base = lvl_keys[0]
            for col in list(levels[base].keys()):
                if not any(col in levels[l] for l in lvl_keys[1:]):
                    final_data[troop][col] = levels[base][col]
                    del levels[base][col]

        # 4b) Flatten troops with exactly one level
        for troop in list(final_data.keys()):
            levels = final_data[troop]
            if isinstance(levels, dict) and len(levels) == 1:
                only_lvl, data_dict = next(iter(levels.items()))
                if isinstance(data_dict, dict):
                    final_data[troop] = data_dict

        # 5) Write final JSON
        with open(f"{save_name}.json", "w", encoding="utf-8") as jf:
            json.dump(final_data, jf, indent=2)

        # 6) Delete the CSV
        if not self.KEEP_CSV:
            try:
                os.remove(file_path)
            except OSError as e:
                logging.warning(f"Could not delete {file_path}: {e}")

    def is_compressed(self, data):
        """Check if data is compressed by looking for compression signatures."""
        # Check for Sig: header (signed compressed data)
        if data[:4] == b"Sig:":
            return True
        # Check for SCLZ (LZHAM)
        if data[:4] == b"SCLZ":
            return True
        # Check for ZSTD magic number
        if len(data) >= 4 and int.from_bytes(data[0:4], byteorder="little") == zstandard.MAGIC_NUMBER:
            return True
        # Check for LZMA (starts with 0x5D)
        if data[0] == 0x5D:
            return True
        # Check for SC header (Supercell compression)
        if data[:2] == b"\x53\x43":
            return True
        # If it looks like plain CSV text, it's not compressed
        if data[:6] == b'"Name"' or data[:6] == b'"name"' or data[:5] == b'"TID"':
            return False
        # Default to compressed if we're not sure
        return True

    def open_file(self, file_path: str) -> dict:
        with open(file_path, "r", encoding="utf-8") as f:
            data: dict = json.load(f)
        return data

    def _translate(self, tid: str):
        self.USED_TIDS.add(tid)
        return self.translation_data.get(tid, {}).get("EN")

    def _parse_upgrade_time(self, level_data: dict) -> int:
        upgrade_time_seconds = (level_data.get("BuildTimeD") or level_data.get("UpgradeTimeD", 0)) * 24 * 60 * 60
        upgrade_time_seconds += (level_data.get("BuildTimeH") or level_data.get("UpgradeTimeH", 0)) * 60 * 60
        upgrade_time_seconds += (level_data.get("BuildTimeM") or level_data.get("UpgradeTimeM", 0)) * 60
        upgrade_time_seconds += level_data.get("BuildTimeS") or level_data.get("UpgradeTimeS", 0)
        return upgrade_time_seconds

    def _parse_resource(self, resource: str) -> str:
        resource_TID: str = self.full_resource_data.get(resource, {}).get("TID")
        return self._translate(resource_TID)

    def _parse_translation_data(self):
        full_translation_data = self.open_file("texts.json")
        other_translations = []
        for language in self.supported_languages:
            with open(f"texts_{language}.json", "r", encoding="utf-8") as f:
                other_translations.append((language, json.load(f)))

        new_translation_data = {}
        for translation_key, translation_data in full_translation_data.items():
            new_translation_data[translation_key] = {"EN": translation_data.get("EN")}
            for lang, language_data in other_translations:
                if not language_data.get(translation_key):
                    continue
                new_translation_data[translation_key][lang.upper()] = language_data.get(translation_key).get(
                    lang.upper())

        self.translation_data = new_translation_data
        return new_translation_data

    def _parse_achievement_data(self):
        self.full_achievement_data = self.open_file("achievements.json")
        new_achievement_data = {}
        for achievement_name, achievement_data in self.full_achievement_data.items():
            tid = achievement_data.get("TID")
            group_map = {0: "home", 1: "builderBase", 2: "clanCapital"}
            if tid not in new_achievement_data:
                new_achievement_data[tid] = {
                    "name": self._translate(achievement_data.get("TID")),
                    "info": self._translate(achievement_data.get("InfoTID")),
                    "completed_message": self._translate(achievement_data.get("CompletedTID")),
                    "TID": {
                        "name": achievement_data.get("TID"),
                        "info": achievement_data.get("InfoTID"),
                        "completed_message": achievement_data.get("CompletedTID"),
                    },
                    "village": group_map.get(achievement_data["UIGroup"]),
                    "ui_priority": achievement_data.get("UIPriority", 0),
                    "levels" : [{
                        "level": achievement_data.get("Level") + 1,
                        "action_count": achievement_data.get("ActionCount"),
                        "action_data": achievement_data.get("ActionData"),
                        "xp": achievement_data.get("ExpReward", 0),
                        "gems": achievement_data.get("DiamondReward", 0)
                    }]
                }
            else:
                new_achievement_data[tid]["levels"].append({
                    "level": achievement_data.get("Level") + 1,
                    "action_count": achievement_data.get("ActionCount"),
                    "action_data": achievement_data.get("ActionData"),
                    "xp": achievement_data.get("ExpReward", 0),
                    "gems": achievement_data.get("DiamondReward", 0)
                })

        return list(new_achievement_data.values())

    def _parse_building_data(self):
        self.full_building_data = self.open_file("buildings.json")
        self.full_supercharges_data = self.open_file("mini_levels.json")
        self.full_townhall_data = self.open_file("townhall_levels.json")
        full_weapon_data: dict = self.open_file("weapons.json")

        new_building_data = []

        #fill in ids, make it easier
        for _id, (building_name, building_data) in enumerate(self.full_building_data.items(), 1000000):
            building_data["_id"] = _id

        for _id, (building_name, building_data) in enumerate(self.full_building_data.items(), 1000000):
            if building_data.get("BuildingClass") in ["Npc", "NonFunctional",
                                                      "Npc Town Hall"] or "Unused" in building_name:
                continue

            village_type = building_data.get("VillageType", 0)
            superchargeable = False

            supercharge_level_data = {}
            for supercharge_data in self.full_supercharges_data.values():
                if supercharge_data.get("TargetBuilding") == building_name:
                    superchargeable = True
                    hold_data = {
                        "upgrade_resource": self._parse_resource(resource=supercharge_data.get("BuildResource")),
                        "levels": []
                    }
                    for level, level_data in supercharge_data.items():
                        if not isinstance(level_data, dict):
                            continue
                        upgrade_time_seconds = self._parse_upgrade_time(level_data)

                        DPS = level_data.get("DPS", 0)
                        # if the level doesnt have a DPS & there is no hitpoints for this row, that means it is a DPS upgrade
                        # unless it is a resource pump, but we dont handle those anyways
                        if not DPS and not level_data.get("Hitpoints"):
                            DPS = supercharge_data.get("DPS", 0)
                        hold_data["levels"].append({
                            "level": int(level),
                            "build_cost": level_data.get("BuildCost"),
                            "build_time": upgrade_time_seconds,
                            "hitpoints_buff": level_data.get("Hitpoints", 0),
                            "dps_buff": DPS,
                        })
                    supercharge_level_data = hold_data
                    break

            #for merged buildings, move the requirement to level 1 since that is when the requirement is actually needed
            if building_data.get("MergeRequirement") is not None:
                building_data["1"]["MergeRequirement"] = building_data.get("MergeRequirement")

            hold_data = {
                "_id": _id,
                "name": self._translate(tid=building_data.get("TID")),
                "info": self._translate(tid=building_data.get("InfoTID")),
                "TID": {
                    "name": building_data.get("TID"),
                    "info": building_data.get("InfoTID"),
                },
                "type": building_data.get("BuildingClass"),
                # we are going to do this purely for builder hut purposes bc lv 1 is gems
                "upgrade_resource": self._parse_resource(
                    resource=building_data.get("BuildResource") or building_data.get("2").get("BuildResource")
                ),
                "village": "home" if not village_type else "builderBase",
                "width": building_data.get("Width", 1), #walls are null for some reason, so let's make it 1
                "superchargeable": superchargeable,
                "levels": []
            }

            # put seasonal defense onto the crafting station
            if _id == 1000097:
                # seasonal defenses are a max townhall thing
                seasonal_def_data = self._parse_seasonal_defense_data()
                hold_data["seasonal_defenses"] = seasonal_def_data

            if building_data.get("GearUpLevelRequirement"):
                hold_data["gear_up"] = {
                    "level_required": building_data.get("GearUpLevelRequirement"),
                    "resource": self._parse_resource(resource=building_data.get("GearUpResource")),
                    "building_id": self.full_building_data.get(building_data.get("GearUpBuilding")).get("_id")
                }


            for level, level_data in building_data.items():
                if not isinstance(level_data, dict):
                    continue

                upgrade_time_seconds = self._parse_upgrade_time(level_data)
                hold_level_data = {
                    "level": level_data.get("BuildingLevel"),
                    "build_cost": level_data.get("BuildCost", 0),
                    "build_time": upgrade_time_seconds,
                    "required_townhall": level_data.get("TownHallLevel"),
                    "hitpoints": level_data.get("Hitpoints", 0),
                    "dps": level_data.get("DPS", 0) or level_data.get("Damage", 0),
                }

                if "AltBuildResource" in level_data:
                    # a wall specific thing since they can use gold + elixir at certain levels
                    hold_level_data["alt_upgrade_resource"] = self._parse_resource(resource=level_data["AltBuildResource"])

                if merge_requirement := level_data.get("MergeRequirement"):
                    merge_list = []
                    buildings = merge_requirement.split(";")
                    for building in buildings:
                        name, level, geared_up = building.split(":")
                        merge_building_data = self.full_building_data.get(name)
                        merge_list.append({
                            "name": self._translate(tid=merge_building_data.get("TID")),
                            "_id": merge_building_data.get("_id"),
                            "geared_up": True if geared_up == "1" else False,
                            "level": int(level)
                        })

                    hold_level_data["merge_requirement"] = merge_list

                if (weapon_name := level_data.get("Weapon")) is not None:

                    weapon_data: dict = full_weapon_data[weapon_name]

                    #if the townhall only has 1 level of weapon, then it is inherently part of the base level,
                    # so just set the dps and continue
                    if weapon_data.get("1") is None:
                        hold_level_data["dps"] = weapon_data.get("DPS")
                    else:
                        hold_weapon_data = {
                            "name": self._translate(tid=weapon_data["TID"]),
                            "info": self._translate(tid=weapon_data["InfoTID"]),
                            "TID": {
                                "name": weapon_data.get("TID"),
                                "info": weapon_data.get("InfoTID"),
                            },
                            "upgrade_resource": self._parse_resource(resource=building_data.get("BuildResource")),
                            "levels": []
                        }
                        for weapon_level, weapon_level_data in weapon_data.items():
                            if not isinstance(weapon_level_data, dict):
                                continue

                            upgrade_time_seconds = self._parse_upgrade_time(weapon_level_data)
                            hold_weapon_data["levels"].append({
                                "level": weapon_level_data.get("Level"),
                                "build_cost": level_data.get("BuildCost"),
                                "build_time": upgrade_time_seconds,
                                "dps": weapon_level_data.get("DPS"),
                            })
                        hold_level_data["weapon"] = hold_weapon_data

                hold_data["levels"].append(hold_level_data)

            if superchargeable:
                #supercharges are always on the last available level
                hold_data["levels"][-1]["supercharge"] = supercharge_level_data
            new_building_data.append(hold_data)

        lab_data = next((item for item in new_building_data if item["name"] == "Laboratory")).get("levels")
        lab_to_townhall = {spot : level_data.get("required_townhall") for spot, level_data in enumerate(lab_data, 1)}
        lab_to_townhall[-1] = 1 # there are troops with no lab ...
        lab_to_townhall[0] = 2
        self.lab_to_townhall = lab_to_townhall

        blacksmith_data = next((item for item in new_building_data if item["name"] == "Blacksmith")).get("levels")
        self.smithy_to_townhall = {spot: level_data.get("required_townhall") for spot, level_data in
                              enumerate(blacksmith_data, 1)}

        pet_house_data = next((item for item in new_building_data if item["name"] == "Pet House")).get("levels")
        self.pethouse_to_townhall = {spot: level_data.get("required_townhall") for spot, level_data in
                                enumerate(pet_house_data, 1)}

        bb_lab_data = next((item for item in new_building_data if item["name"] == "Star Laboratory")).get("levels")
        self.bb_lab_to_townhall = {spot: level_data.get("required_townhall") for spot, level_data in
                              enumerate(bb_lab_data, 1)}

        townhall_unlocks, builderhall_unlocks = self._parse_hall_data()

        for building in new_building_data:
            unlocks = []
            if building["type"] == "Town Hall":
                unlocks = townhall_unlocks
            elif building["type"] == "Town Hall2":
                unlocks = builderhall_unlocks

            for unlock_data in unlocks:
                building["levels"][(unlock_data["level"] - 1)]["unlocks"] = unlock_data["buildings_unlocked"]

        return new_building_data

    def _parse_seasonal_defense_data(self):
        full_seasonal_defenses = self.open_file("seasonal_defense_archetypes.json")
        full_seasonal_modules = self.open_file("seasonal_defense_modules.json")
        full_season_data = self.open_file("seasonal_defense.json")

        seasons = []
        for season_data in full_season_data.values():
            seasons.append(season_data)

        current_season = next((item for item in reversed(seasons) if item.get("TID")), {})
        current_seasonal_defenses: list[str] = [v.get("Archetypes") for k, v in current_season.items() if k.isdigit()]

        for _id, (n, d) in enumerate(full_seasonal_modules.items(), 102000000):
            d["_id"] = _id

        current_max_townhall = int(list(self.full_townhall_data.keys())[-1])
        new_seasonal_defense_data = []
        for _id, (seasonal_def_name, seasonal_def_data) in enumerate(full_seasonal_defenses.items(), 103000000):

            if seasonal_def_name not in current_seasonal_defenses:
                continue

            season_defense_ability = self.full_abilities_data.get(seasonal_def_data.get("SpecialAbility"))

            name_TID = season_defense_ability.get("OverrideTID")
            info_TID = season_defense_ability.get("OverrideInfoTID")

            hold_data = {
                "_id": _id,
                "name": self._translate(tid=name_TID),
                "info": self._translate(tid=info_TID),
                "TID": {
                    "name": name_TID,
                    "info": info_TID,
                },
                "required_townhall": current_max_townhall,
                "modules" : []
            }
            for count, module in enumerate(seasonal_def_data.get("Modules").split(";"), 1):
                module_data = full_seasonal_modules.get(module)

                module_hold_data = {
                    "_id": module_data.get("_id"),
                    "name": self._translate(tid=module_data.get("TID")),
                    "TID": {
                        "name": module_data.get("TID"),
                    },
                    "upgrade_resource": self._parse_resource(module_data.get("BuildResource")),
                    "levels": []
                }
                for level, level_data in module_data.items():
                    if not isinstance(level_data, dict):
                        continue
                    upgrade_time_seconds = self._parse_upgrade_time(level_data)

                    ability_data = self.full_abilities_data.get(module_data.get("SpecialAbility")).get(level)
                    ability_data.pop("ActivateFromGameSystem", None)
                    ability_data.pop("DeactivateFromGameSystem", None)
                    ability_data.pop("Level", None)

                    module_hold_data["levels"].append({
                        "level": int(level),
                        "build_cost": level_data.get("BuildCost"),
                        "build_time": upgrade_time_seconds,
                        "ability_data": ability_data
                    })

                hold_data["modules"].append(module_hold_data)

            new_seasonal_defense_data.append(hold_data)

        return new_seasonal_defense_data

    def _parse_troop_data(self):
        self.full_troop_data = self.open_file("characters.json")
        full_super_troop_data = self.open_file("super_licences.json")
        full_super_troop_data = {v.get("Replacement"): v for k, v in full_super_troop_data.items()}

        name_to_id = {}
        new_troop_data = []
        for _id, (troop_name, troop_data) in enumerate(self.full_troop_data.items(), 4000000):
            if troop_data.get("DisableProduction", False):
                continue
            village_type = troop_data.get("VillageType", 0)
            production_building = self.full_building_data.get(troop_data.get("ProductionBuilding")).get("TID")

            name_to_id[(troop_name, village_type)] = _id
            hold_data = {
                "_id": _id,
                "name": self._translate(tid=troop_data.get("TID")),
                "info": self._translate(tid=troop_data.get("InfoTID")),
                "TID": {
                    "name": troop_data.get("TID"),
                    "info": troop_data.get("InfoTID"),
                },
                "production_building": self._translate(tid=production_building),
                "production_building_level": troop_data.get("BarrackLevel"),
                "upgrade_resource": self._parse_resource(resource=troop_data.get("UpgradeResource")),

                "is_flying": troop_data.get("IsFlying"),
                "is_air_targeting": troop_data.get("AirTargets"),
                "is_ground_targeting": troop_data.get("GroundTargets"),

                "movement_speed": troop_data.get("Speed", 0),

                "attack_speed": troop_data.get("AttackSpeed", 0),
                "attack_range": troop_data.get("AttackRange", 0),
                "housing_space": troop_data.get("HousingSpace"),
                "village": "home" if not village_type else "builderBase",
            }
            is_super_troop = troop_data.get("EnabledBySuperLicence", False)
            is_seasonal_troop = troop_data.get("EnabledByCalendar", False)
            super_troop_data = None
            if is_super_troop:
                super_troop_data = full_super_troop_data.get(troop_name)
                hold_data["super_troop"] = {"original_id": name_to_id[(super_troop_data["Original"], 0)],
                                            "original_min_level": super_troop_data["MinOriginalLevel"]}
            if is_seasonal_troop:
                hold_data["is_seasonal"] = True
            hold_data["levels"] = []

            max_townhall_converter = self.lab_to_townhall
            if troop_data.get("ProductionBuilding") == "Barrack2":
                max_townhall_converter = self.bb_lab_to_townhall

            for level, level_data in troop_data.items():
                if not isinstance(level_data, dict):
                    continue
                #convert times to seconds, all times for all things will be in seconds
                upgrade_time_seconds = self._parse_upgrade_time(level_data)

                if not is_super_troop and not is_seasonal_troop:
                    required_lab_level = level_data.get("LaboratoryLevel")
                    required_townhall = max_townhall_converter[level_data.get("LaboratoryLevel")]
                elif is_super_troop: #for super troops use the original troop's lab level'
                    original_troop = self.full_troop_data.get(super_troop_data["Original"])
                    required_lab_level = original_troop.get(level).get("LaboratoryLevel")
                    required_townhall = max_townhall_converter[required_lab_level]
                elif is_seasonal_troop:
                    required_lab_level = None
                    required_townhall = level_data.get("UpgradeLevelByTH")
                else:
                    continue

                new_level_data = {
                    "level": int(level),
                    "hitpoints": level_data.get("Hitpoints", 0),
                    "dps": level_data.get("DPS", 0),

                    "upgrade_time": upgrade_time_seconds,
                    "upgrade_cost": level_data.get("UpgradeCost", 0),
                    "required_lab_level": required_lab_level,
                    "required_townhall": required_townhall,
                }
                hold_data["levels"].append(new_level_data)

            if not hold_data["levels"]:
                continue
            new_troop_data.append(hold_data)

        return new_troop_data

    def _parse_guardian_data(self):
        full_guardian_data = self.open_file("guardians.json")

        new_guardian_data = []
        for _id, (guardian_name, guardian_data) in enumerate(full_guardian_data.items(), 107000000):
            if guardian_data.get("Deprecated", False):
                continue
            character_data = self.full_troop_data.get(guardian_data.get("CharacterDatas"))
            hold_data = {
                "_id": _id,
                "name": self._translate(tid=guardian_data.get("TID")),
                "info": self._translate(tid=guardian_data.get("InfoTID")),
                "TID": {
                    "name": guardian_data.get("TID"),
                    "info": guardian_data.get("InfoTID"),
                },
                "upgrade_resource": self._parse_resource(resource=character_data.get("UpgradeResource")),

                "is_flying": character_data.get("IsFlying"),
                "is_air_targeting": character_data.get("AirTargets"),
                "is_ground_targeting": character_data.get("GroundTargets"),

                "movement_speed": character_data.get("Speed"),

                "attack_speed": character_data.get("AttackSpeed"),
                "attack_range": character_data.get("AttackRange"),
                "levels": []
            }

            for level, level_data in character_data.items():
                if not isinstance(level_data, dict):
                    continue
                #convert times to seconds, all times for all things will be in seconds
                upgrade_time_seconds = self._parse_upgrade_time(level_data)

                #hard coded for now, didn't find where this is defined, except "HousesGuardians" on the Townhall data
                required_townhall = 18

                new_level_data = {
                    "level": int(level),
                    "hitpoints": level_data.get("Hitpoints"),
                    "dps": level_data.get("DPS"),

                    "upgrade_time": upgrade_time_seconds,
                    "upgrade_cost": level_data.get("UpgradeCost", 0),
                    "required_townhall": required_townhall,
                }
                hold_data["levels"].append(new_level_data)

            new_guardian_data.append(hold_data)

        return new_guardian_data

    def _parse_spell_data(self):
        full_spell_data = self.open_file("spells.json")

        new_spell_data = []
        for _id, (spell_name, spell_data) in enumerate(full_spell_data.items(), 26000000):
            if spell_data.get("DisableProduction", False):
                continue

            production_building = self.full_building_data.get(spell_data.get("ProductionBuilding")).get("TID")
            hold_data = {
                "_id": _id,
                "name": self._translate(spell_data.get("TID")),
                "info": self._translate(spell_data.get("InfoTID")),
                "TID": {
                    "name": spell_data.get("TID"),
                    "info": spell_data.get("InfoTID"),
                },
                "production_building": self._translate(production_building),
                "production_building_level": spell_data.get("SpellForgeLevel"),
                "upgrade_resource": self._parse_resource(resource=spell_data.get("UpgradeResource")),
                "housing_space": spell_data.get("HousingSpace"),
            }
            is_seasonal_spell = spell_data.get("EnabledByCalendar", False)
            if is_seasonal_spell:
                hold_data["is_seasonal"] = is_seasonal_spell
            hold_data["levels"] = []

            for level, level_data in spell_data.items():
                if not isinstance(level_data, dict):
                    continue
                # convert times to seconds, all times for all things will be in seconds
                upgrade_time_seconds = self._parse_upgrade_time(level_data)

                duration_ms = level_data.get("NumberOfHits", 0) * level_data.get("TimeBetweenHitsMS", 0)
                radius = spell_data.get("Radius", 0) or level_data.get("Radius", 0)
                new_level_data = {
                    "level": int(level),
                    "duration": int(duration_ms / 1000),
                    "radius": round(radius / 100, 1),
                    "damage": level_data.get("Damage", 0) or level_data.get("PoisonDPS", 0),
                    "upgrade_time": upgrade_time_seconds,
                    "upgrade_cost": level_data.get("UpgradeCost", 0),
                    "required_lab_level": level_data.get("LaboratoryLevel"),
                    "required_townhall": level_data.get("UpgradeLevelByTH") or self.lab_to_townhall[level_data.get("LaboratoryLevel")],
                }
                hold_data["levels"].append(new_level_data)

            if not hold_data["levels"]:
                continue
            new_spell_data.append(hold_data)

        return new_spell_data

    def _parse_hero_data(self):
        self.full_hero_data = self.open_file("heroes.json")

        new_hero_data = []
        for _id, (hero_name, hero_data) in enumerate(self.full_hero_data.items(), 28000000):
            village_type = hero_data.get("VillageType", 0)
            hold_data = {
                "_id": _id,
                "name": self._translate(tid=hero_data.get("TID")),
                "info": self._translate(hero_data.get("InfoTID")),
                "TID": {
                    "name": hero_data.get("TID"),
                    "info": hero_data.get("InfoTID"),
                },
                "production_building": self._translate(tid="TID_HERO_TAVERN") if not village_type else None,
                "production_building_level": hero_data.get("1", {}).get("RequiredHeroTavernLevel"),
                "upgrade_resource": self._parse_resource(resource=hero_data.get("UpgradeResource")),

                "is_flying": hero_data.get("IsFlying"),
                "is_air_targeting": hero_data.get("AirTargets"),
                "is_ground_targeting": hero_data.get("GroundTargets"),

                "movement_speed": hero_data.get("Speed"),
                "attack_speed": hero_data.get("AttackSpeed"),
                "attack_range": hero_data.get("AttackRange"),
                "village": "home" if not village_type else "builderBase",
                "levels": []
            }

            for level, level_data in hero_data.items():
                if not isinstance(level_data, dict):
                    continue
                # convert times to seconds, all times for all things will be in seconds
                upgrade_time_seconds = self._parse_upgrade_time(level_data)

                new_level_data = {
                    "level": int(level),
                    "hitpoints": level_data.get("Hitpoints"),
                    "dps": level_data.get("DPS"),

                    "upgrade_time": upgrade_time_seconds,
                    "upgrade_cost": level_data.get("UpgradeCost", 0),

                    "required_townhall": level_data.get("RequiredTownHallLevel"),
                    "required_hero_tavern_level": level_data.get("RequiredHeroTavernLevel"),
                }
                hold_data["levels"].append(new_level_data)

            new_hero_data.append(hold_data)

        return new_hero_data

    def _parse_pet_data(self):
        full_pet_data = self.open_file("pets.json")

        new_pet_data = []
        for _id, (pet_name, pet_data) in enumerate(full_pet_data.items(), 73000000):
            if pet_data.get("Deprecated", False) or pet_name in ["PhoenixEgg"]:
                continue

            hold_data = {
                "_id": _id,
                "name": self._translate(tid=pet_data.get("TID")),
                "info": self._translate(tid=pet_data.get("InfoTID")),
                "TID": {
                    "name": pet_data.get("TID"),
                    "info": pet_data.get("InfoTID"),
                },
                "production_building": self._translate(tid="TID_PET_SHOP"),
                "production_building_level": pet_data.get("1").get("LaboratoryLevel"),
                "upgrade_resource": self._parse_resource('DarkElixir'),

                "is_flying": pet_data.get("IsFlying"),
                "is_air_targeting": pet_data.get("AirTargets"),
                "is_ground_targeting": pet_data.get("GroundTargets"),

                "movement_speed": pet_data.get("Speed"),
                "attack_speed": pet_data.get("AttackSpeed"),
                "attack_range": pet_data.get("AttackRange"),
                "levels": []
            }

            for level, level_data in pet_data.items():
                if not isinstance(level_data, dict):
                    continue
                # convert times to seconds, all times for all things will be in seconds
                upgrade_time_seconds = self._parse_upgrade_time(level_data)

                new_level_data = {
                    "level": int(level),
                    "hitpoints": level_data.get("Hitpoints"),
                    "dps": level_data.get("DPS"),

                    "upgrade_time": upgrade_time_seconds,
                    "upgrade_cost": level_data.get("UpgradeCost", 0),
                    "required_pet_house_level": level_data.get("LaboratoryLevel"),
                    "required_townhall": self.pethouse_to_townhall[level_data.get("LaboratoryLevel")],
                }
                hold_data["levels"].append(new_level_data)

            new_pet_data.append(hold_data)

        return new_pet_data

    def _parse_equipment_data(self):
        full_equipment_data = self.open_file("character_items.json")

        new_equipment_data = []
        for _id, (equipment_name, equipment_data) in enumerate(full_equipment_data.items(), 90000000):
            if equipment_data.get("Deprecated", False):
                continue

            main_abilities = equipment_data.get("MainAbilities").split(";")
            extra_abilities = equipment_data.get("ExtraAbilities", "").split(";")
            hero_TID = self.full_hero_data.get(equipment_data.get("AllowedCharacters").split(";")[0]).get("TID")
            hold_data = {
                "_id": _id,
                "name": self._translate(tid=equipment_data.get("TID")),
                "info": self._translate(tid=equipment_data.get("InfoTID")),
                "TID": {
                    "name": equipment_data.get("TID"),
                    "info": equipment_data.get("InfoTID"),
                    "production_building": "TID_SMITHY",
                },
                "production_building": self._translate(tid="TID_SMITHY"),
                "production_building_level": equipment_data.get("1").get("RequiredBlacksmithLevel"),
                "rarity": equipment_data.get("Rarity"),
                "hero": self._translate(tid=hero_TID),
                "levels": []
            }

            for level, level_data in equipment_data.items():
                if not isinstance(level_data, dict):
                    continue

                shiny_ore = 0
                glowy_ore = 0
                starry_ore = 0
                upgrade_resources = level_data.get("UpgradeResources", "").split(";")
                upgrade_costs = str(level_data.get("UpgradeCosts", "")).split(";")

                if upgrade_costs[0] != "":
                    for resource, cost in zip(upgrade_resources, upgrade_costs):
                        resource = resource.strip()
                        cost = int(cost)
                        if resource == "CommonOre":
                            shiny_ore += cost
                        elif resource == "RareOre":
                            glowy_ore += cost
                        elif resource == "EpicOre":
                            starry_ore += cost

                new_level_data = {
                    "level": int(level),
                    "hitpoints": level_data.get("Hitpoints", 0),
                    "dps": level_data.get("DPS", 0),
                    "heal_on_activation" : level_data.get("HealOnActivation", 0),
                    "required_blacksmith_level": level_data.get("RequiredBlacksmithLevel"),
                    "required_townhall": self.smithy_to_townhall[level_data.get("RequiredBlacksmithLevel")],
                    "upgrade_cost": {
                        "shiny_ore": shiny_ore,
                        "glowy_ore": glowy_ore,
                        "starry_ore": starry_ore
                    }
                }

                main_ability_levels = str(level_data.get("MainAbilityLevels", "")).split(";")

                if main_ability_levels[0] != "":
                    main_ability_json = []
                    for main_ability, main_ability_level in zip(main_abilities, main_ability_levels):
                        full_ability = self.full_abilities_data.get(main_ability)
                        ability = full_ability.get(main_ability_level)
                        ability["name"] = self._translate(tid=full_ability.get("TID"))
                        ability["info"] = self._translate(tid=full_ability.get("InfoTID"))
                        main_ability_json.append(ability)

                    if main_ability_json:
                        new_level_data["abilities"] = main_ability_json

                extra_ability_levels = str(level_data.get("ExtraAbilityLevels", "")).split(";")
                if extra_ability_levels[0] != "":
                    extra_ability_json = []
                    for extra_ability, extra_ability_level in zip(extra_abilities, extra_ability_levels):
                        full_ability = self.full_abilities_data.get(extra_ability)
                        ability = full_ability.get(extra_ability_level)
                        if ability:
                            ability["name"] = self._translate(tid=full_ability.get("TID"))
                            extra_ability_json.append(ability)

                    if extra_ability_json:
                        new_level_data["abilities"].extend(extra_ability_json)

                hold_data["levels"].append(new_level_data)

            new_equipment_data.append(hold_data)

        return new_equipment_data

    def _parse_trap_data(self):
        full_trap_data = self.open_file("traps.json")

        new_trap_data = []
        for _id, (trap_name, trap_data) in enumerate(full_trap_data.items(), 12000000):
            if trap_data.get("Disabled", False) or trap_data.get("EnabledByCalendar", False):
                continue
            village_type = trap_data.get("VillageType", 0)

            hold_data = {
                "_id": _id,
                "name": self._translate(tid=trap_data.get("TID")),
                "info": self._translate(tid=trap_data.get("InfoTID")),
                "TID" : {
                    "name" : trap_data.get("TID"),
                    "info" : trap_data.get("InfoTID"),
                },
                "width": trap_data.get("Width"),
                "air_trigger": trap_data.get("AirTrigger", False),
                "ground_trigger": trap_data.get("GroundTrigger", False),
                "damage_radius": trap_data.get("DamageRadius"),
                "trigger_radius": trap_data.get("TriggerRadius"),
                "village": "home" if not village_type else "builderBase",

                "upgrade_resource": self._parse_resource(resource=trap_data.get("BuildResource")),
                "levels" : []
            }
            for level, level_data in trap_data.items():
                if not isinstance(level_data, dict):
                    continue

                upgrade_time_seconds = self._parse_upgrade_time(level_data)

                hold_data["levels"].append({
                    "level": int(level),
                    "build_cost": level_data.get("BuildCost"),
                    "build_time": upgrade_time_seconds,
                    "required_townhall": level_data.get("TownHallLevel"),
                    "damage": level_data.get("Damage", 0),
                })

            new_trap_data.append(hold_data)

        return new_trap_data

    def _parse_decoration_data(self):
        full_deco_data = self.open_file("decos.json")
        new_deco_data = []
        for _id, (deco_name, deco_data) in enumerate(full_deco_data.items(), 18000000):
            if deco_data.get("TID") in ["TID_DECORATION_GENERIC", "TID_DECORATION_NATIONAL_FLAG"]:
                continue
            village_type = deco_data.get("VillageType", 0)
            hold_data = {
                "_id": _id,
                "name": self._translate(deco_data.get("TID")),
                "TID": {
                    "name": deco_data.get("TID"),
                },
                "width": deco_data.get("Width"),
                "not_in_shop": deco_data.get("NotInShop", False),
                "pass_reward": deco_data.get("BPReward", False),
                "max_count": deco_data.get("MaxCount", 1),
                "build_resource": self._parse_resource(resource=deco_data.get("BuildResource")),
                "build_cost": deco_data.get("BuildCost"),
                "village": "home" if not village_type else "builderBase"
            }
            new_deco_data.append(hold_data)

        return new_deco_data

    def _parse_capital_part_data(self):
        full_capital_part_data = self.open_file("building_parts.json")
        new_capital_part_data = []
        for _id, (part_name, part_data) in enumerate(full_capital_part_data.items(), 82000000):
            if part_data.get("Deprecated", False):
                continue
            name = part_name.replace("PlayerHouse_", "").replace("_", " ").title()
            nums = 0
            for phrase in ["01", "02", "03", "04", "05", "06", "07", "08", "09", "10"]:
                if phrase in name:
                    name = name.replace(phrase, "")
                    nums = int(phrase)
            name = name.split(" ", 1)
            name = f"{name[1]} {name[0]}"
            if nums:
                name = f"{name} {nums}"

            # make it match the API enums
            type_mapping = {"Deco" : "decoration"}
            slot_type = type_mapping.get(part_data.get("LayoutSlot"), part_data.get("LayoutSlot").lower())
            new_capital_part_data.append({
                "_id": _id,
                "name": name.title(),
                "slot_type": slot_type,
                "pass_reward": part_data.get("BattlePassReward", False),
            })

        return new_capital_part_data

    def _parse_obstacle_data(self):
        full_obstacle_data = self.open_file("obstacles.json")

        new_obstacle_data = []
        for _id, (obstacle_name, obstacle_data) in enumerate(full_obstacle_data.items(), 8000000):
            village_type = obstacle_data.get("VillageType", 0)
            hold_data = {
                "_id": _id,
                "name": self._translate(tid=obstacle_data.get("TID")),
                "TID": {
                    "name": obstacle_data.get("TID"),
                },
                "width": obstacle_data.get("Width"),
                "clear_resource": self._parse_resource(resource=obstacle_data.get("ClearResource")),
                "clear_cost": obstacle_data.get("ClearCost"),
                "loot_resource": self._parse_resource(resource=obstacle_data.get("LootResource")),
                "loot_count": obstacle_data.get("LootCount"),
                "village": "home" if not village_type else "builderBase"
            }
            new_obstacle_data.append(hold_data)

        return new_obstacle_data

    def _parse_scenery_data(self):
        full_scenery_data = self.open_file("village_backgrounds.json")

        new_scenery_data = []
        for _id, (scenery_name, scenery_data) in enumerate(full_scenery_data.items(), 60000000):
            type_map = {
                "WAR" : "war",
                "BB" : "builderBase",
                "HOME" : "home"
            }
            if scenery_data.get("HomeType") not in type_map:
                continue

            if not self._translate(tid=scenery_data.get("TID")):
                continue

            hold_data = {
                "_id": _id,
                "name": self._translate(tid=scenery_data.get("TID")),
                "TID": {
                    "name": scenery_data.get("TID"),
                },
                "type": type_map.get(scenery_data.get("HomeType")),
                "music" : scenery_data.get("Music"),
            }
            if  scenery_data.get("FreeBackground", False):
                scenery_data["free"] = True
            if scenery_data.get("DefaultBackground", False):
                scenery_data["default"] = True

            new_scenery_data.append(hold_data)

        return new_scenery_data

    def _parse_skin_data(self):
        full_skin_data = self.open_file("skins.json")

        new_skins_data = []
        for _id, (skin_name, skin_data) in enumerate(full_skin_data.items(), 52000000):
            character = skin_data.get("character") or skin_data.get("Character")
            if not skin_data.get("TID") or character not in self.full_hero_data.keys() or not skin_data.get("Tier"):
                continue
            hold_data = {
                "_id": _id,
                "name": self._translate(tid=skin_data.get("TID")),
                "TID": {
                    "name": skin_data.get("TID"),
                },
                "tier": skin_data.get("Tier").title(),
                "character" : character,
            }
            new_skins_data.append(hold_data)

        return new_skins_data

    def _parse_helper_data(self):
        full_helper_data = self.open_file("villager_apprentices.json")

        new_helper_data = []
        for _id, (helper_name, helper_data) in enumerate(full_helper_data.items(), 93000000):
            hold_data = {
                "_id": _id,
                "name": self._translate(tid=helper_data.get("TID")),
                "info": self._translate(tid=helper_data.get("InfoTID")),
                "TID": {
                    "name": helper_data.get("TID"),
                    "info": helper_data.get("InfoTID"),
                },
                "upgrade_resource": self._parse_resource(resource=helper_data.get("CostResource")),
                "levels": []
            }
            for level, level_data in helper_data.items():
                if not isinstance(level_data, dict):
                    continue
                hold_data["levels"].append({
                    "level": int(level),
                    "required_townhall": level_data.get("RequiredTownHallLevel"),
                    "upgrade_cost": level_data.get("Cost"),
                    "boost_time_seconds": level_data.get("BoostTimeSeconds"),
                    "boost_multiplier": level_data.get("BoostMultiplier"),
                })

            new_helper_data.append(hold_data)

        return new_helper_data

    def _parse_war_league_data(self):
        full_war_league_data = self.open_file("war_leagues.json")

        new_war_league_data = []
        for _id, (war_league_name, war_league_data) in enumerate(full_war_league_data.items(), 48000000):
            if not war_league_data.get("Name"): #skip Unranked, no data
                continue
            new_war_league_data.append({
                "_id": _id,
                "name": self._translate(tid=war_league_data.get("TID")),
                "TID": {
                    "name": war_league_data.get("TID"),
                },
                "cwl_medals": {
                    "first_place": war_league_data.get("LeagueWinReward"),
                    "position_medal_diff": war_league_data.get("LeaguePosRewardEffect"),
                    "bonus_reward": war_league_data.get("BonusMedalReward"),
                    "minimum_bonus_amount": war_league_data.get("MinNumMedalBonuses"),
                },
                "promotions": war_league_data.get("NumPromotions"),
                "demotions": war_league_data.get("NumDemotions"),
                "15v15_only": war_league_data.get("AllowFirstWarSizeOnly")
            })

        return new_war_league_data

    def _parse_league_tier_data(self):
        full_league_tier_data = self.open_file("league_tiers.json")

        new_league_tier_data = []
        for _id, (league_name, league_data) in enumerate(full_league_tier_data.items(), 105000000):
            league_tier = _id - 105000000
            hold_data = {
                "_id": _id,
                "name": self._translate(tid=league_data.get("TID")),
                "league_tier": league_tier,
                "TID": {
                    "name": league_data.get("TID"),
                },
                "group_size": league_data.get("GroupSizeMax"),
                "demote_percentage": league_data.get("DemotePercentage"),
                "promote_percentage": league_data.get("PromotePercentage"),
                "battle_count": league_data.get("MaxBattleCount"),
                "trophy_start": league_data.get("TrophyFloor"),
                "clan_score": league_data.get("TopClanScore"),
                "townhall_cap": None,
                "rewards" : []
            }
            highest_townhall = 0
            rewards = []
            for tier, level_data in league_data.items():
                if not isinstance(level_data, dict):
                    continue
                if tier == "1": #always empty idky
                    continue
                townhall_level = level_data.get("TH")
                th_min_league_tier = self.full_townhall_data.get(str(townhall_level)).get("LeagueTier", 0)
                if league_tier < th_min_league_tier and league_tier != 0:
                    continue
                highest_townhall = max(townhall_level, highest_townhall)
                rewards.append({
                    "townhall_level": townhall_level,
                    "resources": {
                        "gold": level_data.get("GoldReward"),
                        "elixir": level_data.get("ElixirReward"),
                        "dark_elixir": level_data.get("DarkElixirReward")
                    },
                    "star_bonus": {
                        "gold": level_data.get("GoldRewardStarBonus"),
                        "elixir": level_data.get("ElixirRewardStarBonus"),
                        "dark_elixir": level_data.get("DarkElixirRewardStarBonus"),
                        "shiny_ore": level_data.get("CommonOreRewardStarBonus"),
                        "glowy_ore": level_data.get("RareOreRewardStarBonus"),
                        "starry_ore": level_data.get("EpicOreRewardStarBonus")
                    }
                })
            hold_data["townhall_cap"] = highest_townhall
            hold_data["rewards"] = rewards
            new_league_tier_data.append(hold_data)

        return new_league_tier_data

    def _parse_hall_data(self):
        builderhall_data = []
        townhall_data = []
        id_quantity_map = {}

        for _id, (hall_level, hall_data) in enumerate(self.full_townhall_data.items(), 1):
            builderhall_unlocks = []
            townhall_unlocks = []
            for field, data in hall_data.items():
                building_data = self.full_building_data.get(field)
                if not building_data:
                    continue
                village_type = building_data.get("VillageType", 0)
                id = building_data.get("_id")
                quantity = data

                current_quantity = id_quantity_map.get(id, 0)
                new_quantity = quantity - current_quantity

                if not new_quantity:
                    continue

                if id not in id_quantity_map:
                    id_quantity_map[id] = new_quantity
                else:
                    id_quantity_map[id] += new_quantity

                if not village_type: #is home village
                    townhall_unlocks.append({
                        "name": self._translate(tid=building_data.get("TID")),
                        "_id": id,
                        "quantity": new_quantity
                    })
                else:
                    builderhall_unlocks.append({
                        "name": self._translate(tid=building_data.get("TID")),
                        "_id": id,
                        "quantity": new_quantity
                    })
            townhall_data.append({"level": _id, "buildings_unlocked": townhall_unlocks})
            if builderhall_unlocks:
                builderhall_data.append({"level": _id, "buildings_unlocked": builderhall_unlocks})

        return townhall_data, builderhall_data

    def create_master_json(self):
        self._parse_translation_data()

        self.full_abilities_data = self.open_file("special_abilities.json")
        self.full_resource_data = self.open_file("resources.json")

        master_data = {
            "buildings": self._parse_building_data(),
            "traps" : self._parse_trap_data(),
            "troops": self._parse_troop_data(),
            "guardians": self._parse_guardian_data(),
            "spells" : self._parse_spell_data(),
            "heroes": self._parse_hero_data(),
            "pets": self._parse_pet_data(),
            "equipment": self._parse_equipment_data(),
            "decorations": self._parse_decoration_data(),
            "obstacles": self._parse_obstacle_data(),
            "sceneries": self._parse_scenery_data(),
            "skins": self._parse_skin_data(),
            "capital_house_parts": self._parse_capital_part_data(),
            "helpers": self._parse_helper_data(),
            "war_leagues": self._parse_war_league_data(),
            "league_tiers": self._parse_league_tier_data(),
            "achievements": self._parse_achievement_data(),
        }
        with open(f"{self.BASE_PATH}static_data.json", "w", encoding="utf-8") as jf:
            jf.write(json.dumps(master_data, indent=2))

        if self.PRUNE_TRANSLATIONS:
            for key in list(self.translation_data.keys()):
                if key not in self.USED_TIDS:
                    del self.translation_data[key]

        with open(f"{self.BASE_PATH}translations.json", "w", encoding="utf-8") as jf:
            jf.write(json.dumps(self.translation_data, indent=2))

        for file_path in self.TARGETS:
            # 6) Delete the extra jsons
            if self.KEEP_JSON:
                continue
            try:
                file_path = file_path.replace("csv", "json")
                os.remove(file_path)
            except OSError as e:
                logging.warning(f"Could not delete {file_path}: {e}")

    def generate_constants(self):
        static_data = self.open_file("static_data.json")

        troops = static_data["troops"]
        spells = static_data["spells"]
        heroes = static_data["heroes"]
        equipment = static_data["equipment"]
        pets = static_data["pets"]
        buildings = static_data["buildings"]
        achievements = static_data["achievements"]

        lists_to_write = {
            'ELIXIR_TROOP_ORDER':  [
                t["name"] for t in troops
                if t["production_building"] == "Barracks"
                and not t.get("is_seasonal", False)
                and not "super_troop" in t
            ],
            'DARK_ELIXIR_TROOP_ORDER': [
                t["name"] for t in troops
                if t["production_building"] == "Dark Barracks"
                and not t.get("is_seasonal", False)
                and not "super_troop" in t
            ],
            'HV_TROOP_ORDER': 'ELIXIR_TROOP_ORDER + DARK_ELIXIR_TROOP_ORDER',
            'SIEGE_MACHINE_ORDER': [t["name"] for t in troops if t["production_building"] == "Workshop"],
            'SUPER_TROOP_ORDER': [t["name"] for t in troops if "super_troop" in t],
            'HOME_TROOP_ORDER': 'HV_TROOP_ORDER + SIEGE_MACHINE_ORDER',
            'SEASONAL_TROOP_ORDER': [t["name"] for t in troops if t.get("is_seasonal", False)],
            'BUILDER_TROOPS_ORDER': [t["name"] for t in troops if t["village"] == "builderBase"],
            'ELIXIR_SPELL_ORDER': [
                s["name"] for s in spells
                if s["upgrade_resource"] == "Elixir"
                and not s.get("is_seasonal", False)
            ],
            'DARK_ELIXIR_SPELL_ORDER': [s["name"] for s in spells if s["upgrade_resource"] == "Dark Elixir"],
            'SEASONAL_SPELL_ORDER': [s["name"] for s in spells if s.get("is_seasonal", False)],
            'SPELL_ORDER': 'ELIXIR_SPELL_ORDER + DARK_ELIXIR_SPELL_ORDER',
            'HOME_BASE_HERO_ORDER': [
                h["name"] for h in sorted(heroes, key=lambda x: x["levels"][0]["required_townhall"])
                if h["village"] == "home"
            ],
            'BUILDER_BASE_HERO_ORDER': [h["name"] for h in heroes if h["village"] == "builderBase"],
            'HERO_ORDER': 'HOME_BASE_HERO_ORDER + BUILDER_BASE_HERO_ORDER',
            'PETS_ORDER': [p["name"] for p in pets],
            'EQUIPMENT': [e["name"] for e in equipment],
            'HV_BUILDINGS': [b["name"] for b in buildings if b["village"] == "home"],
            'ACHIEVEMENT_ORDER': [
                a["name"] for a in sorted(achievements,
                key=lambda x: ({'home': 0, 'builderBase': 1, 'clanCapital': 2}.get(
                x["village"], 0), -x["ui_priority"]))
            ], # same order as in-game
        }
        constants_path = Path(__file__).parent.parent / "constants.py"

        with open(constants_path, 'w') as f:
            f.write('"""Auto-generated constants from static game data."""\n\n')
            for name, lst in lists_to_write.items():
                if isinstance(lst, str):
                    f.write(f"{name} = {lst}\n\n")
                else:
                    # Manual formatting: each item on its own line
                    f.write(f"{name} = [\n")
                    for item in lst:
                        f.write(f"    {repr(item)},\n")
                    f.write("]\n\n")
        
        print(f"Constants written to {constants_path}")

    async def download_files(self):
        if not self.FINGERPRINT:
            self.FINGERPRINT = await self.get_fingerprint()

        BASE_URL = f"https://game-assets.clashofclans.com/{self.FINGERPRINT}"

        fingerprint_file = await self.download(url=f"{BASE_URL}/fingerprint.json", as_json=True)

        for file_data in fingerprint_file.get("files"):
            file_path: str = file_data["file"]
            if not file_path.startswith("logic/") and not file_path.startswith("localization/"):
                continue

            download_url = f"{BASE_URL}/{file_path}"
            print(f"Downloading: {download_url}")
            data = await self.download(url=download_url)

            save_path = file_path.split("/", )[-1]

            if file_path.startswith("localization/") and "texts" not in save_path:
                self.supported_languages.append(save_path.replace(".csv", ""))
                save_path = f"texts_{save_path}"

            with open(save_path, "wb") as f:
                f.write(data)

            print(f"Processing: {save_path}")
            self.process_csv(data=data, file_path=save_path, save_name=save_path.split(".")[0])
            self.TARGETS.append(save_path)

        self.create_master_json()

    def run(self):
        asyncio.run(self.download_files())

if __name__ == "__main__":
    StaticUpdater().run()




import unittest
import orjson
import coc

from coc.utils import UnitStatList
from coc.enums import HV_TROOP_ORDER
from coc.troop import TROOPS_FILE_PATH

import tracemalloc

tracemalloc.start()

class ClashMetaTests(unittest.TestCase):

    # def test_load_never(self):
    #     client = coc.login_with_keys("", load_game_data=coc.LoadGameData(never=True))
    #     with self.assertRaises(RuntimeError) as cm:
    #         client.parse_army_link("")
    #
    #     self.assertEqual(str(cm.exception), "Troop and Spell game metadata must be loaded to use this feature.")

    def test_load_default(self):
        client = coc.Client(load_game_data=coc.LoadGameData(default=True))
        client._create_holders()

        for holder in (client._troop_holder, client._spell_holder, client._hero_holder, client._pet_holder):
            self.assertTrue(holder.loaded)

    def test_load_startup_only(self):
        client = coc.Client(load_game_data=coc.LoadGameData(startup_only=True))
        client._create_holders()

        for holder in (client._troop_holder, client._spell_holder, client._hero_holder, client._pet_holder):
            self.assertTrue(holder.loaded)

        troop = client.get_troop("Barbarian")
        self.assertIs(type(troop), type(coc.Troop))

    def test_troop(self):
        client = coc.Client()
        client._create_holders()
        barb = client.get_troop("Barbarian")
        self.assertIs(type(barb), type(coc.Troop))

        self.assertEqual(barb.name, "Barbarian")
        self.assertIsInstance(barb.level, UnitStatList)


class TroopMeta(unittest.TestCase):
    def setUp(self) -> None:
        self.client = coc.Client()
        self.client._create_holders()
        
        ## load the characters data
        with open(TROOPS_FILE_PATH, "rb") as f:
            self.raw_troops = orjson.loads(f.read())
        
        
        self.troop = self.client.get_troop("Barbarian")
        # self.barb_i = self.client.get_troop("Barbarian", level=12)

    def test_types(self):
        self.assertIs(type(self.troop), type(coc.Troop))

    def test_levels(self):
        self.assertIsInstance(self.troop.level, UnitStatList)

        with self.assertRaises(IndexError):
            self.troop.level[0]

        self.assertEqual(len(self.troop.level), len(self.troop.lab_level))
    
    def test_troops(self):
        for troop_name in HV_TROOP_ORDER:
            troop = self.client.get_troop(troop_name)
            self.assertIs(type(troop), type(coc.Troop))
            self.assertEqual(troop.name, troop_name)
            self.assertTrue(troop.is_home_base )
            # get the raw troop data
            troop_data = self.raw_troops.get(troop.internal_name)
            self.assertEqual(troop.is_home_base and troop.is_elixir_troop, troop_data.get("ProductionBuilding") == "Barrack")
            self.assertEqual(troop.is_home_base and troop.is_dark_troop, troop_data.get("ProductionBuilding") == "Dark Elixir Barrack")
            self.assertEqual(troop.ground_target, troop_data.get("GroundTargets"))
            max_level = 0
            for k, v in troop_data.items():
                if not isinstance(v, dict):
                    continue
                level = v.get("VisualLevel")
                if not level:
                    continue
                max_level = max(max_level, level)
                level_troop = self.client.get_troop(troop_name, level=level)
                self.assertEqual(level_troop.name, troop_name, f"{level_troop.name} name is wrong")
                total_seconds = 0
                total_seconds += v.get("UpgradeTimeD", 0) * 24 * 60 * 60
                total_seconds += v.get("UpgradeTimeH", 0) * 60 * 60
                total_seconds += v.get("UpgradeTimeM", 0) * 60
                total_seconds += v.get("UpgradeTimeS", 0)
                self.assertEqual(level_troop.upgrade_cost or 0,
                                 v.get("UpgradeCost", 0) or 0,
                                 f"{level_troop.name} upgrade cost level {level} is wrong")
                self.assertEqual(level_troop.upgrade_time.total_seconds(), total_seconds,
                                 f"{level_troop.name} upgrade time level {level} is wrong")
            self.assertEqual(troop.max_level, max_level, f"{troop.name} max level is wrong")




import typing

from .abc import BasePlayer
from .war_attack import WarAttack

if typing.TYPE_CHECKING:
    from .wars import ClanWar
    from .war_clans import WarClan


class ClanWarMember(BasePlayer):
    """Represents a War Member that the API returns.
    Depending on which method calls this, some attributes may
    be ``None``.

    Attributes
    ----------
    tag: :class:`str`
        The player's tag
    name: :class:`str`
        The player's name
    town_hall:
        :class:`int`: The member's townhall level.
    map_position:
        :class:`int`: The member's map position in the war.
    defense_count:
        :class:`int`: The number of times this member has been attacked.
    war:
        :class:`ClanWar`: The current war this member is in.
    clan:
        :class:`WarClan`: The member's clan.
    attack_cls: :class:`coc.WarAttack`
        The type which the attacks in :attr:`ClanWarMember.attacks` will be of.
        Ensure any overriding of this inherits from :class:`coc.WarAttack`.
    """

    __slots__ = (
        "tag",
        "name",
        "town_hall",
        "defense_count",
        "__iter_attacks",
        "_attack_opponents",
        "_best_opponent_attacker",
        "map_position",
        "war",
        "clan",
        "attack_cls",
        "_client",
        "_attacks",
    )

    def __init__(self, *, data, client, war, clan, **kwargs):
        super().__init__(data=data, client=client)
        self._client = client
        self._attacks = []
        self.war = war  # type: ClanWar
        self.clan = clan  # type: WarClan
        self.attack_cls = kwargs.pop('attack_cls', WarAttack)
        self._from_data(data)

    def _from_data(self, data):
        data_get = data.get

        self.name: str = data_get("name")
        self.tag: str = data_get("tag")
        self.town_hall: int = data_get("townhallLevel")
        self.map_position: int = data_get("mapPosition")
        self.defense_count: int = data_get("opponentAttacks")

        self.__iter_attacks = (
            self.attack_cls(data=adata, client=self._client, war=self.war) for adata in data_get("attacks", [])
        )

        self._best_opponent_attacker: str = data_get("bestOpponentAttack", {}).get("attackerTag")

    @property
    def best_opponent_attack(self) -> WarAttack:
        """:class:`WarAttack`: Returns the best opponent attack on this base."""
        return self.war.get_attack(self._best_opponent_attacker, self.tag)

    @property
    def previous_best_opponent_attack(self):
        """:class:`WarAttack`: Returns the previous best opponent attack on this base.

        This is useful for calculating the new stars and/or destruction for new attacks.
        """

        def key(item):
            # 100^3 > 99^2 > 50^2 > 99^1. Smallest/largest values possible.
            return item != self.best_opponent_attack and item.destruction ** item.stars

        # Potential caveat: how does order effect this?
        return max(self.defenses, key=key)

    @property
    def attacks(self) -> typing.List[WarAttack]:
        """List[:class:`WarAttack`]: The member's attacks this war. Could be an empty list."""
        list_attacks = self._attacks
        if list_attacks:
            return list_attacks

        list_attacks = self._attacks = list(self.__iter_attacks)
        return list_attacks

    @property
    def defenses(self) -> typing.List[WarAttack]:
        """List[:class:`WarAttack`]: The member's defenses this war. Could be an empty list."""
        return self.war.get_defenses(self.tag)

    @property
    def is_opponent(self) -> bool:
        """:class:`bool`: Indicates whether the member is from the opponent clan or not."""
        return self.clan and self.war.opponent and self.clan.tag == self.war.opponent.tag or False

    @property
    def star_count(self):
        """:class:`int`: Get the total number of stars the member has gained this war."""
        return sum(attack.stars for attack in self.attacks)


class ClanWarLeagueClanMember(BasePlayer):
    """Represents a clan member who is a part of the Clan War League master roster.

    Attributes
    ----------
    tag: :class:`str`
        The player's tag
    name: :class:`str`
        The player's name
    town_hall: :class:`int`
        The player's town hall level.
    """

    __slots__ = ("town_hall",)

    def __init__(self, *, data, client):
        super().__init__(data=data, client=client)
        self.town_hall = data.get("townHallLevel")



import asyncio
import unittest
import os
import coc

# Siege data was grabbed from:
# https://clashofclans.fandom.com/wiki/Workshop
MACHINES = {
    "Siege Barracks":
        {
            "release": 13,
            "max_release_level": 4,
            "max_at": [
                {
                    "th": 13,
                    "max": 4
                }
            ]
        },
    "Wall Wrecker":
        {
            "release": 12,
            "max_release_level": 3,
            "max_at": [
                {
                    "th": 13,
                    "max": 4
                }
            ]
        },

}


class TestTroopLevel(unittest.IsolatedAsyncioTestCase):
    def setUp(self) -> None:
        self.coc_client = coc.Client()
        self.coc_client._create_holders()
        self.machines = MACHINES

    def test_siege_machines(self):
        for machine_key in self.machines.keys():
            machine = self.machines.get(machine_key)

            machine_obj = self.coc_client.get_troop(
                name=machine_key,
                townhall=machine.get("release"))
            max_level = machine_obj.get_max_level_for_townhall(12)
            self.assertEqual(None if 12 < machine.get('release') else machine.get('max_release_level'), max_level)


if __name__ == '__main__':
    unittest.main()




from typing import Optional, List, TYPE_CHECKING

from .miscmodels import BaseLeague, PlayerHouseElement, try_enum, Achievement, Label, League, LegendStatistics
from .enums import (
    Role,
)

from .constants import (
    HERO_ORDER,
    BUILDER_TROOPS_ORDER,
    HOME_TROOP_ORDER,
    SPELL_ORDER,
    SIEGE_MACHINE_ORDER,
    ACHIEVEMENT_ORDER,
    SUPER_TROOP_ORDER,
    PETS_ORDER,
    EQUIPMENT,
)
from .abc import BasePlayer
from .hero import Hero, Equipment, Pet
from .player_clan import PlayerClan
from .spell import Spell
from .troop import Troop
from .utils import cached_property


if TYPE_CHECKING:
    # pylint: disable=cyclic-import
    from .clans import Clan  # noqa
    from .client import Client


class ClanMember(BasePlayer):
    """Represents a Clash of Clans Clan Member.

    Attributes
    ----------
    tag: :class:`str`
        The player's tag
    name: :class:`str`
        The player's name
    clan: Optional[:class:`Clan`]
        The player's clan. If the player is clanless, this will be ``None``.
    role: :class:`Role`
        The member's role in a clan. To get a string as rendered in-game, do ``str(member.role)``.
    exp_level: :class:`int`
        The member's experience level.
    league: :class:`League`
        The member's current league.
    builder_base_league: :class:`BaseLeague`
        The member's current builder base league.
    trophies: :class:`int`
        The member's trophy count.
    builder_base_trophies: :class:`int`
        The member's builder base trophy count.
    town_hall: :class:`int`
        The player's town hall level. In case the player hasn't logged in since 2019, this will be `0`.
    clan_rank: :class:`int`
        The member's rank in the clan.
    clan_previous_rank: :class:`int`
        The member's rank before the last leaderboard change
        (ie if Bob overtakes Jim in trophies, and they switch ranks on the leaderboard,
        and you want to find out their previous rankings, this will help.).
    builder_base_rank: :class:`int`
        The member's rank in the clan based on builder base trophies.
    donations: :class:`int`
        The member's donation count for this season.
    received: :class:`int`
        The member's donations received count for this season.
    clan_cls: :class:`coc.Clan`
        The class to use to create the :attr:`ClanMember.clan` attribute.
        Ensure any overriding of this inherits from :class:`coc.Clan` or :class:`coc.PlayerClan`.
    league_cls: :class:`coc.League`
        The class to use to create the :attr:`ClanMember.league` attribute.
        Ensure any overriding of this inherits from :class:`coc.League`.
    builder_base_league_cls: Optional[:class:`coc.League`]
        The class to use to create the :attr:`ClanMember.builder_base_league` attribute.
        Ensure any overriding of this inherits from :class:`coc.BaseLeague`.
    """

    __slots__ = (
        "_client",
        "clan",
        "role",
        "exp_level",
        "league",
        "builder_base_league",
        "trophies",
        "builder_base_trophies",
        "clan_rank",
        "clan_previous_rank",
        "builder_base_rank",
        "donations",
        "received",
        "clan_cls",
        "league_cls",
        "builder_base_league_cls",
        "_player_house_elements",
        "player_house_element_cls",
        "_iter_player_house_elements",
        "_cs_player_house_elements",
        "town_hall",
    )

    def __init__(self, *, data, client, clan=None, **_):
        super().__init__(data=data, client=client)
        self._client = client
        self.clan_cls = PlayerClan
        self.league_cls = League
        self.builder_base_league_cls = BaseLeague
        self.player_house_element_cls = PlayerHouseElement

        self._from_data(data)
        if clan:
            self.clan = clan

    def _from_data(self, data: dict) -> None:
        data_get = data.get

        self.exp_level: int = data_get("expLevel")
        self.trophies: int = data_get("trophies")
        self.builder_base_trophies: int = data_get("builderBaseTrophies")
        self.clan_rank: int = data_get("clanRank")
        self.clan_previous_rank: int = data_get("previousClanRank")
        self.builder_base_rank: int = data.get("builderBaseRank")
        self.donations: int = data_get("donations")
        self.received: int = data_get("donationsReceived")
        player_house_element_cls = self.player_house_element_cls
        self.clan = try_enum(self.clan_cls, data=data_get("clan"), client=self._client)
        self.league = try_enum(self.league_cls, data=data_get("leagueTier"), client=self._client)

        # fall back if no league given, this is what in-game shows
        self.builder_base_league = try_enum(self.builder_base_league_cls,
                                            data=data_get(
                                                "builderBaseLeague",
                                                {"id":44000000,"name":"Wood League V"}
                                            ),
                                            client=self._client)
        self.role = data_get("role") and Role(value=data["role"])
        self.town_hall: int = data_get("townHallLevel")
        self._iter_player_house_elements = (player_house_element_cls(data=adata)
                                            for adata in data_get("playerHouse", {}).get("elements", []))

    async def get_detailed_clan(self) -> Optional["Clan"]:
        """Get clan details for the player's clan. If the player's clan is ``None``,this will return ``None``.

        Example
        ---------

        .. code-block:: python3

            player = await client.get_player('tag')
            clan = await player.get_detailed_clan()
        """
        return self.clan and await self._client.get_clan(self.clan.tag)

    @cached_property("_cs_player_house_elements")
    def player_house_elements(self) -> List[PlayerHouseElement]:
        """List[:class:`PlayerHouseElement`]: A list of :class:`PlayerHouseElement` objects that the player has."""
        return list(self._iter_player_house_elements)


class RankedPlayer(ClanMember):
    """
    Represents a leaderboard-ranked player.

    Attributes
    ----------
    attack_wins: :class:`int`
        The player's number of attack wins. If retrieving info for builder base leader-boards, this will be ``None``.
    defense_wins: :class:`int`
        The player's number of defense wins. If retrieving info for builder base leader-boards, this will be ``None``.
    builder_base_trophies: :class:`int`
        The player's builder base trophy count. If retrieving info for regular leader-boards, this will be ``None``.
    rank: :class:`int`
        The player's rank in the clan leaderboard.
    previous_rank: :class:`int`
        The member's rank before the last clan leaderboard change
        (ie if Bob overtakes Jim in trophies, and they switch ranks on the leaderboard,
        and you want to find out their previous rankings, this will help.).
    """

    __slots__ = ("attack_wins", "defense_wins", "builder_base_trophies", "rank", "previous_rank")

    def _from_data(self, data: dict) -> None:
        super()._from_data(data)

        data_get = data.get
        self.attack_wins: int = data_get("attackWins")
        self.defense_wins: int = data_get("defenseWins")
        self.builder_base_trophies: int = data_get("builderBaseTrophies")
        self.rank: int = data_get("rank")
        self.previous_rank: int = data_get("previousRank")


class Player(ClanMember):
    """
    Represents a Clash of Clans Player.

    Attributes
    ----------
    achievement_cls: :class:`Achievement`
        The constructor used to create the :attr:`Player.achievements` list.
        This must inherit from :class:`Achievement`.
    hero_cls: :class:`Hero`
        The constructor used to create the :attr:`Player.heroes` list. This must inherit from :class:`Hero`.
    label_cls: :class:`Label`
        The constructor used to create the :attr:`Player.labels` list. This must inherit from :class:`Label`.
    spell_cls: :class:`Spell`
        The constructor used to create the :attr:`Player.spells` list. This must inherit from :class:`Spell`.
    troop_cls: :class:`Troop`
        The constructor used to create the :attr:`Player.troops` list. This must inherit from :class:`Troop`.
    equipment_cls: :class:`Equipment`
        The constructor used to create the :attr:`Player.equipment` list. This must inherit from :class:`Equipment`.
    attack_wins: :class:`int`
        The number of attacks the player has won this season.
    defense_wins: :class:`int`
        The number of defenses the player has won this season.
    best_trophies: :class:`int`
        The player's best recorded trophies for the home base.
    war_stars: :class:`int`
        The player's total war stars.
    town_hall_weapon: Optional[:class:`int`]
        The player's town hall weapon level, or ``None`` if it doesn't exist.
    builder_hall: :class:`int`
        The player's builder hall level, or 0 if it hasn't been unlocked.
    best_builder_base_trophies: :class:`int`
        The player's best builder base trophy count.
    clan_capital_contributions: :class:`int`
        The player's total contribution to clan capitals
    legend_statistics: Optional[:class:`LegendStatistics`]
        The player's legend statistics, or ``None`` if they have never been in the legend league.
    war_opted_in: Optional[:class:`bool`]
        Whether the player has selected that they are opted "in" (True) for wars, or opted "out" (False).
        This will be ``None`` if the player is not in a clan.
    """

    __slots__ = (
        "clan",
        "attack_wins",
        "defense_wins",
        "best_trophies",
        "war_stars",
        "town_hall_weapon",
        "builder_hall",
        "best_builder_base_trophies",
        "clan_capital_contributions",
        "legend_statistics",
        "war_opted_in",
        "_achievements",
        "_heroes",
        "_pets",
        "_equipment",
        "_labels",
        "_spells",
        "_home_troops",
        "_builder_troops",
        "_super_troops",
        "achievement_cls",
        "hero_cls",
        "label_cls",
        "spell_cls",
        "troop_cls",
        "pet_cls",
        "equipment_cls",

        "_iter_achievements",
        "_iter_heroes",
        "_iter_labels",
        "_iter_spells",
        "_iter_troops",
        "_iter_pets",
        "_iter_equipment",

        "_cs_labels",
        "_cs_achievements",
        "_cs_troops",
        "_cs_home_troops",
        "_cs_builder_troops",
        "_cs_siege_machines",
        "_cs_pets",
        "_cs_super_troops",
        "_cs_heroes",
        "_cs_spells",
        "_cs_equipment",

        "_game_files_loaded",
        "_load_game_data",

    )

    def __init__(self, *, data, client: 'Client', load_game_data=None, **_):
        self._client = client

        self._achievements = None  # type: Optional[dict]
        self._heroes = None  # type: Optional[dict]
        self._spells = None  # type: Optional[dict]
        self._home_troops: dict = {}
        self._builder_troops: dict = {}
        self._super_troops: list = []
        self._pets = None  # type: Optional[dict]
        self._equipment: Optional[dict] = None

        self.achievement_cls = Achievement
        self.hero_cls = Hero
        self.label_cls = Label
        self.spell_cls = Spell
        self.troop_cls = Troop
        self.pet_cls = Pet
        self.equipment_cls = Equipment


        if load_game_data is not None:
            self._load_game_data = load_game_data
        elif self._client and self._client.load_game_data.never:
            self._load_game_data = False
        else:
            self._load_game_data = True

        if self._load_game_data and client is None:
            raise RuntimeError("Client must be provided if load_game_data is True.")

        super().__init__(data=data, client=client)

    def _from_data(self, data: dict) -> None:
        super()._from_data(data)
        data_get = data.get

        self.attack_wins: int = data_get("attackWins")
        self.defense_wins: int = data_get("defenseWins")
        self.best_trophies: int = data_get("bestTrophies")
        self.war_stars: int = data_get("warStars")
        self.town_hall_weapon: int = data_get("townHallWeaponLevel")
        self.builder_hall: int = data_get("builderHallLevel", 0)
        self.best_builder_base_trophies: int = data_get("bestBuilderBaseTrophies")
        self.clan_capital_contributions: int = data_get("clanCapitalContributions")
        self.legend_statistics = try_enum(LegendStatistics, data=data_get("legendStatistics"))

        try:
            self.war_opted_in: Optional[bool] = True if data["warPreference"] == "in" else False
        except KeyError:
            # not in a clan / war preference not there
            self.war_opted_in: Optional[bool] = None

        label_cls = self.label_cls
        achievement_cls = self.achievement_cls

        pet_lookup = PETS_ORDER

        self._iter_labels = (label_cls(data=ldata, client=self._client) for ldata in data_get("labels", []))
        self._iter_achievements = (achievement_cls(data=adata) for adata in data_get("achievements", []))

        self._iter_troops = (
            self.troop_cls(
                client=self._client,
                data=tdata,
                static_data=self._client._get_static_data(
                    item_name=tdata["name"],
                    village=tdata["village"],
                    section="troops",
                    bypass=not self._load_game_data
                ) if self._client else None
            ) for tdata in data_get("troops", []) if tdata["name"] not in pet_lookup
        )

        self._iter_heroes = (
            self.hero_cls(
                client=self._client,
                data=hdata,
                static_data=self._client._get_static_data(
                    item_name=hdata["name"],
                    village=hdata["village"],
                    section="heroes",
                    bypass=not self._load_game_data
                ) if self._client else None
            ) for hdata in data_get("heroes", [])
        )

        self._iter_spells = (
            self.spell_cls(
                client=self._client,
                data=sdata,
                static_data=self._client._get_static_data(
                    item_name=sdata["name"],
                    section="spells",
                    bypass=not self._load_game_data
                ) if self._client else None
            ) for sdata in data_get("spells", [])
        )

        self._iter_pets = (
            self.pet_cls(
                client=self._client,
                data=tdata,
                static_data=self._client._get_static_data(
                    item_name=tdata["name"],
                    section="pets",
                    bypass=not self._load_game_data
                ) if self._client else None
            ) for tdata in data_get("troops", []) if tdata["name"] in pet_lookup
        )

        self._iter_equipment = (
            self.equipment_cls(
                client=self._client,
                data=edata,
                static_data=self._client._get_static_data(
                    item_name=edata["name"],
                    section="equipment",
                    bypass=not self._load_game_data
                ) if self._client else None
            ) for edata in data_get('heroEquipment', [])
        )

    def _inject_clan_member(self, member):
        if member:
            self.clan_rank = getattr(member, "clan_rank", None)
            self.clan_previous_rank = getattr(member, "clan_previous_rank", None)


    @cached_property("_cs_labels")
    def labels(self) -> List[Label]:
        """List[:class:`Label`]: A list of :class:`Label` objects that the player has."""
        return list(self._iter_labels)

    @cached_property("_cs_achievements")
    def achievements(self) -> List[Achievement]:
        """List[:class:`Achievement`]: A list of the player's achievements."""
        # at the time of writing, the API presents achievements in the order
        # added to the game which doesn't match in-game order.
        achievement_dict = {a.name: a for a in self._iter_achievements}
        sorted_achievements = {}
        for name in ACHIEVEMENT_ORDER:
            try:
                sorted_achievements[name] = achievement_dict[name]
            except KeyError:
                continue

        self._achievements = sorted_achievements
        return list(sorted_achievements.values())

    def get_achievement(self, name: str, default_value=None) -> Optional[Achievement]:
        """Gets an achievement with the given name.

        Parameters
        -----------
        name: :class:`str`
            The name of an achievement as found in-game.
        default_value
            The value to return if the ``name`` is not found. Defaults to ``None``.

        Returns
        --------
        Optional[:class:`Achievement`]
            The returned achievement or the ``default_value`` if not found, which defaults to ``None``.
        """
        if not self._achievements:
            _ = self.achievements

        try:
            return self._achievements[name]
        except KeyError:
            return default_value

    @cached_property("_cs_troops")
    def troops(self) -> List[Troop]:
        """List[:class:`Troop`]: A :class:`List` of the player's :class:`Troop`.

        Troops are **not** ordered in this attribute. Use either :attr:`Player.home_troops`
        or :attr:`Player.builder_troops` if you want an ordered list.

        This includes:
        - Elixir Troops (Barbarian, Balloon, etc.)
        - Dark Elixir Troops (Minion, Hog Rider, etc.)
        - Siege Machines (Log Launcher, etc.)
        - **Boosted** Super Troops
        - Builder Troops (Raged Barbarian, etc.)

        This **does not** include:
        - Heroes
        - Hero Pets
        - Spells
        - Un-boosted Super Troops
        """
        troops = []

        for troop in self._iter_troops:
            if (self._load_game_data and troop.is_super_troop) or troop.name in SUPER_TROOP_ORDER:
                self._super_troops.append(troop)
                if troop.is_active:
                    self._home_troops[troop.name] = troop
                    troops.append(troop)

            elif troop.is_home_base:
                self._home_troops[troop.name] = troop
                troops.append(troop)
            elif troop.is_builder_base:
                self._builder_troops[troop.name] = troop
                troops.append(troop)

        return troops

    @cached_property("_cs_home_troops")
    def home_troops(self) -> List[Troop]:
        """List[:class:`Troop`]: A :class:`List` of the player's home-base :class:`Troop`.

        This will return troops in the order found in both barracks and laboratory in-game.

        This includes:
        - Elixir Troops (Barbarian, Balloon, etc.)
        - Dark Elixir Troops (Minion, Hog Rider, etc.)
        - Siege Machines (Log Launcher, etc.)
        - **Boosted** Super Troops
        """
        order = {k: v for v, k in enumerate(HOME_TROOP_ORDER)}

        if not self._home_troops:
            _ = self.troops

        return list(sorted(self._home_troops.values(), key=lambda t: order.get(t.name, 0)))

    @cached_property("_cs_builder_troops")
    def builder_troops(self) -> List[Troop]:
        """List[:class:`Troop`]: A :class:`List` of the player's builder-base :class:`Troop`.

        This will return troops in the order found in both barracks and laboratory in-game.

        This includes:
        - Builder troops
        """
        order = {k: v for v, k in enumerate(BUILDER_TROOPS_ORDER)}

        if not self._builder_troops:
            _ = self.troops

        return list(sorted(self._builder_troops.values(), key=lambda t: order.get(t.name, 0)))

    @cached_property("_cs_siege_machines")
    def siege_machines(self) -> List[Troop]:
        """List[:class:`Troop`]: A :class:`List` of the player's siege-machine :class:`Troop`.

        This will return siege machines in the order found in both barracks and laboratory in-game.

        This includes:
        - Siege machines only.
        """
        order = {k: v for v, k in enumerate(SIEGE_MACHINE_ORDER)}
        troops = (t for t in self.troops if t.name in SIEGE_MACHINE_ORDER or t.is_siege_machine)
        return list(sorted(troops, key=lambda t: order.get(t.name, 0)))

    @cached_property("_cs_pets")
    def pets(self) -> List[Pet]:
        """List[:class:`Pet`]: A :class:`List` of the player's hero pets.

        This will return hero pets in the order found in the Pet House in-game.

        This includes:
        - Hero pets only.
        """
        order = {k: v for v, k in enumerate(PETS_ORDER)}
        pets = list(sorted(self._iter_pets, key=lambda t: order.get(t.name, 0)))
        self._pets = {p.name: p for p in pets}
        return pets

    def get_pet(self, name: str, default_value=None) -> Optional[Pet]:
        """Gets the pet with the given name.

        Parameters
        -----------
        name: :class:`str`
            The name of a pet as found in-game.
        default_value:
            The default value to return if a pet with ``name`` is not found. Defaults to ``None``.

        Returns
        --------
        Optional[:class:`Pet`]
            The returned pet or the ``default_value`` if not found, which defaults to ``None``.

        """
        if not self._pets:
            _ = self.pets

        try:
            return self._pets[name]
        except KeyError:
            return default_value

    @cached_property('_cs_equipment')
    def equipment(self) -> List[Equipment]:
        """List[:class:`Equipment`]: A :class:`List` of the player's hero equipment.

        This will return hero equipment in the order it is in the player's profile

        This includes:
        - Hero equipment only.
        """
        order = {k: v for v, k in enumerate(EQUIPMENT)}
        equipment = list(sorted(self._iter_equipment, key=lambda t: order.get(t.name, 0)))
        self._equipment = {e.name: e for e in equipment}
        return equipment

    def get_equipment(self, name: str, default_value=None) -> Optional[Equipment]:
        """Gets the hero equipment with the given name.

        Parameters
        -----------
        name: :class:`str`
            The name of a hero equipment as found in-game.
        default_value:
            The default value to return if a hero equipment with ``name`` is not found. Defaults to ``None``.

        Returns
        --------
        Optional[:class:`Equipment`]
            The returned hero equipment or the ``default_value`` if not found, which defaults to ``None``.
        """
        if not self._equipment:
            _ = self.equipment

        try:
            return self._equipment[name]
        except KeyError:
            return default_value

    @cached_property("_cs_super_troops")
    def super_troops(self) -> List[Troop]:
        """List[:class:`Troop`]: A :class:`List` of the player's super troops.

        This will return super troops in the order found in the super troop boosting building, in game.

        This includes:
        - All super troops, boosted or not.
        """
        order = {k: v for v, k in enumerate(SUPER_TROOP_ORDER)}

        if not self._super_troops:
            _ = self.troops

        return list(sorted(self._super_troops, key=lambda t: order.get(t.name, 0)))

    def get_troop(self, name: str, is_home_troop=None, default_value=None) -> Optional[Troop]:
        """Gets a troop with the given name.

        Parameters
        -----------
        name: :class:`str`
            The name of a troop as found in-game.
        is_home_troop: :class:`bool`
            Whether the troop you're trying to find is a home troop. This changes how the lookup is done,
            in order to facilitate searching for a ``Baby Dragon``. By default, this will search from both
            builder and home troops.
        default_value
            The value to return if the ``name`` is not found. Defaults to ``None``.

        Returns
        --------
        Optional[:class:`Troop`]
            The returned troop or the ``default_value`` if not found, which defaults to ``None``.
        """
        _ = self.troops

        if is_home_troop is None:
            lookup = {**self._builder_troops, **self._home_troops}
        elif is_home_troop is True:
            lookup = self._home_troops
        elif is_home_troop is False:
            lookup = self._builder_troops
        else:
            raise TypeError("is_home_troop must be of type bool not {0!r}".format(is_home_troop))

        try:
            return lookup[name]
        except KeyError:
            return default_value

    @cached_property("_cs_heroes")
    def heroes(self) -> List[Hero]:
        """List[:class:`Hero`]: A :class:`List` of the player's :class:`Hero`.

        This will return heroes in the order found in the store and laboratory in-game.
        """
        heroes_dict = {h.name: h for h in self._iter_heroes}
        sorted_heroes = {}
        for name in HERO_ORDER:
            # have to do it this way because it's less expensive than removing None's if they don't have a troop.
            try:
                sorted_heroes[name] = heroes_dict[name]
            except KeyError:
                continue

        self._heroes = sorted_heroes
        return list(sorted_heroes.values())

    def get_hero(self, name: str, default_value=None) -> Optional[Hero]:
        """Gets the hero with the given name.

        Parameters
        -----------
        name: :class:`str`
            The name of a hero as found in-game.
        default_value:
            The default value to return if a hero with ``name`` is not found. Defaults to ``None``.

        Returns
        --------
        Optional[:class:`Hero`]
            The returned hero or the ``default_value`` if not found, which defaults to ``None``.

        """
        if not self._heroes:
            _ = self.heroes

        try:
            return self._heroes[name]
        except KeyError:
            return default_value

    @cached_property("_cs_spells")
    def spells(self) -> List[Spell]:
        """List[:class:`Spell`]: A :class:`List` of the player's :class:`Spell` ordered as they appear in-game.

        This will return spells in the order found in both spell factory and laboratory in-game.
        """
        self._spells = {s.name: s for s in self._iter_spells}
        dict_spells = self._spells
        order = {k: v for v, k in enumerate(SPELL_ORDER)}

        return list(sorted(
                dict_spells.values(),
                key=lambda s: order.get(s.name, 0)))

    def get_spell(self, name: str, default_value=None) -> Optional[Spell]:
        """Gets the spell with the given name.

        Parameters
        -----------
        name: :class:`str`
            The name of a spell as found in-game.
        default_value:
            The default value to return if a spell with ``name`` is not found. Defaults to ``None``.

        Returns
        --------
        Optional[:class:`Spell`]
            The returned spell or the ``default_value`` if not found, which defaults to ``None``.
        """
        if not self._spells:
            _ = self.spells

        try:
            return self._spells[name]
        except KeyError:
            return default_value




from .abc import LeveledUnit
from .enums import Resource, VillageType, ProductionBuildingType
from .miscmodels import TimeDelta, TID

class Spell(LeveledUnit):
    """Represents a Spell object as returned by the API, optionally filled with game data.

    Attributes
    ----------
    name: :class:`str`
        The spell's name.
    village: :class:`VillageType`
        The village type (home or builder base) where this spell belongs.
    max_level: :class:`int`
        The maximum level this spell can be upgraded to.
    id: :class:`int`
        The spell's unique identifier.
    info: :class:`str`
        Description of the spell.
    TID: :class:`TID`
        The spell's translation IDs for localization.
    production_building: :class:`ProductionBuildingType`
        The building type that produces this spell.
    production_building_level: :class:`int`
        The required level of the production building to unlock this spell.
    upgrade_resource: :class:`Resource`
        The resource type required to upgrade this spell.
    housing_space: :class:`int`
        The housing space this spell occupies.
    is_seasonal: :class:`bool`
        Whether this is a seasonal/temporary spell.
    duration: :class:`TimeDelta`
        The duration of the spell's effect.
    damage: :class:`int`
        The damage dealt by the spell.
    radius: :class:`float`
        The spell's area of effect radius.
    upgrade_time: :class:`TimeDelta`
        The time required to upgrade to this level.
    upgrade_cost: :class:`int`
        The cost to upgrade to this level.
    required_lab_level: :class:`int`
        The laboratory level required to upgrade to the next level.
    required_townhall: :class:`int`
        The townhall level required to upgrade to the next level.
    """

    __slots__ = (
        "name",
        "village",
        "max_level",
        "id",
        "info",
        "TID",
        "production_building",
        "production_building_level",
        "upgrade_resource",
        "housing_space",
        "is_seasonal",
        "duration",
        "damage",
        "radius",
        "upgrade_time",
        "upgrade_cost",
        "required_lab_level",
        "required_townhall",
        "_raw_data"
    )

    def __init__(self, data: dict, static_data: dict | None, level: int = 0, client = None):
        super().__init__(
            initial_level=data.get("level") or level,
            static_data=static_data
        )

        if client and client.raw_attribute and data:
            self._raw_data = data

        if data:
            self.name: str = data["name"]
            self.village = VillageType(value=data["village"])
            self.max_level: int = data["maxLevel"]

        if static_data:
            self.id: int = static_data["_id"]
            self.name: str = static_data["name"]
            self.info: str = static_data["info"]
            self.TID: TID = TID(data=static_data["TID"])

            self.production_building = ProductionBuildingType(value=static_data["production_building"])
            self.production_building_level: int = static_data["production_building_level"]
            self.upgrade_resource = Resource(value=static_data["upgrade_resource"])

            self.housing_space: int = static_data["housing_space"]

            self.is_seasonal: bool = static_data.get("is_seasonal", False)

            self.village = VillageType.home

            self._load_level_data()

    def _load_level_data(self):
        if not self._static_data:
            return

        level_data = self._static_data["levels"][self._level - 1]

        self.duration = TimeDelta(seconds=level_data["duration"])
        self.damage: int = level_data["damage"]
        self.radius: float = level_data["radius"]
        self.upgrade_time = TimeDelta(seconds=level_data["upgrade_time"])
        self.upgrade_cost: int = level_data["upgrade_cost"]
        self.required_lab_level: int = level_data["required_lab_level"]
        self.required_townhall: int = level_data["required_townhall"]



from enum import Enum


class ExtendedEnum(Enum):
    """An Enum class that allows for the `__str__` method to be implemented."""

    def __str__(self):
        try:
            return self.in_game_name
        except NotImplementedError:
            return self.value

    def __eq__(self, other):
        """Check if the enum is equal to another enum or a string."""
        if isinstance(other, Enum):
            return self.value == other.value
        elif isinstance(other, str):
            return str(self.name) == other or str(self.value) == other
        return False

    @property
    def in_game_name(self) -> str:
        raise NotImplementedError

    @classmethod
    def values(cls):
        return list(map(lambda c: c.value, cls))

    @classmethod
    def names(cls):
        return list(map(lambda c: c.name, cls))


class PlayerHouseElementType(ExtendedEnum):
    """Enum to map the type of element of the player house."""

    ground = "ground"
    roof = "roof"
    foot = "foot"  # API docs say this exists, but unable to find it anywhere. Looks like this means `walls`
    deco = "decoration"
    walls = "walls"

    @property
    def in_game_name(self) -> str:
        """Get a neat client-facing string value for the element type."""
        lookup = {"ground": "Ground", "roof": "Roof", "foot": "Foot", "decoration": "Decoration", "walls": "Walls"}
        return lookup[self.value]


class Role(ExtendedEnum):
    """Enum to map a player's role in the clan."""

    member = "member"
    elder = "admin"
    co_leader = "coLeader"
    leader = "leader"

    @property
    def in_game_name(self) -> str:
        """Get a neat client-facing string value for the role."""
        lookup = {"member": "Member", "admin": "Elder", "coLeader": "Co-Leader", "leader": "Leader"}
        return lookup[self.value]


class WarRound(ExtendedEnum):
    previous_war = 0
    current_war = 1
    current_preparation = 2

    def __str__(self):
        return self.name

    @property
    def in_game_name(self) -> str:
        lookup = ["Previous War", "Current War", "Current Preparation"]
        return lookup[self.value]


class BattleModifier(ExtendedEnum):
    """Enum to map the type of battle modifiers."""
    none = "none"
    hard_mode = "hardMode"

    @property
    def in_game_name(self) -> str:
        """Get a neat client-facing string value for the battle modifier."""
        lookup = {"none": "None", "hardMode": "Hard Mode"}
        return lookup[self.value]


class WarState(ExtendedEnum):
    """Enum to map the state of the war.
    Compared to the api docs a few states are missing, but those were never observed in the wild."""
    not_in_war = "notInWar"
    preparation = "preparation"
    in_war = "inWar"
    war_ended = "warEnded"

    @property
    def in_game_name(self) -> str:
        """Get a neat client-facing string value for the war state."""
        lookup = {"notInWar": "Not in War", "preparation": "Preparation", "inWar": "In War", "warEnded": "War Ended"}
        return lookup[self.value]


class WarResult(ExtendedEnum):
    """Enum to map the result of the war"""
    win = "win"
    lose = "lose"
    tie = "tie"

    @property
    def in_game_name(self) -> str:
        """Get a neat client-facing string value for the war state."""
        lookup = {"win": "Win", "lose": "Lose", "tie": "Tie"}
        return lookup[self.value]


class ClanType(ExtendedEnum):
    open = "open"
    closed = "closed"
    invite_only = "inviteOnly"

    @property
    def in_game_name(self) -> str:
        """Get a neat client-facing string value for the clan type."""
        lookup = {"open": "Open", "closed": "Closed", "inviteOnly": "Invite Only"}
        return lookup[self.value]

class ProductionBuildingType(ExtendedEnum):
    barracks = "Barracks"
    dark_barracks = "Dark Barracks"
    spell_factory = "Spell Factory"
    dark_spell_factory = "Dark Spell Factory"
    hero_hall = "Hero Hall"
    builder_barracks = "Builder Barracks"
    blacksmith = "Blacksmith"
    pet_house = "Pet House"
    workshop = "Workshop"
    none = "None"


class Resource(ExtendedEnum):
    gold = "Gold"
    elixir = "Elixir"
    dark_elixir = "Dark Elixir"
    builder_gold = "Builder Gold"
    builder_elixir = "Builder Elixir"
    gems = "Gems"
    shiny_ore = "Shiny Ore"
    glowy_ore = "Glowy Ore"
    starry_ore = "Starry Ore"
    medal_event_soft_currency = "Medal Event Soft Currency"
    league_medals = "League Medals"
    sparky_stones = "Sparky Stones"


class BuildingType(ExtendedEnum):
    army = "Army"
    town_hall = "Town Hall"
    town_hall_weapon = "Town Hall Weapon"
    builder_hall = "Town Hall2"
    resource = "Resource"
    wall = "Wall"
    defense = "Defense"
    worker = "Worker"
    worker_2 = "Worker2"
    helper = "Helper"


class VillageType(ExtendedEnum):
    home = "home"
    builder_base = "builderBase"
    clan_capital = "clanCapital"

class SceneryType(ExtendedEnum):
    home = "home"
    builder_base = "builderBase"
    war = "war"

class EquipmentRarity(ExtendedEnum):
    common = "Common"
    epic = "Epic"


class SkinTier(ExtendedEnum):
    default = "Default"
    standard = "Basic"
    gold = "Gold"
    legendary = "Legendary"







import typing

if typing.TYPE_CHECKING:
    # pylint: disable=cyclic-import
    from .war_members import ClanWarMember


class WarAttack:
    """Represents a Clash of Clans War Attack

    Attributes
    ----------
    war:
        :class:`ClanWar` - The war this attack belongs to
    stars:
        :class:`int` - The stars achieved
    destruction:
        :class:`float` - The destruction achieved as a percentage (of 100)
    order:
        :class:`int` - The attack order in this war
    attacker_tag:
        :class:`str` - The attacker tag
    defender_tag:
        :class:`str` - The defender tag
    duration:
        :class:`int` - Duration of attack in seconds
    """

    __slots__ = ("war",
                 "member",
                 "stars",
                 "destruction",
                 "order",
                 "attacker_tag",
                 "defender_tag",
                 "duration",
                 "_raw_data",
                 "_client")

    def __repr__(self):
        attrs = [
            ("war", repr(self.war)),
            ("attacker_tag", self.attacker_tag),
            ("defender_tag", self.defender_tag),
            ("stars", self.stars),
            ("destruction", self.destruction),
            ("order", self.order),
            ("duration", self.duration),
        ]
        return "<%s %s>" % (self.__class__.__name__, " ".join("%s=%r" % t for t in attrs),)

    def __hash__(self):
        return hash(self.attacker_tag) << self.order

    def __eq__(self, other):
        # potential future bug: because I'm comparing attacker tag, defender tag and order,
        # if the same 2 clans go into war again in the future,
        # and the same people attack each other at the same order in war and
        # you're trying to compare attacks from the first war (stored in db) and the future war (live)
        # then it's gonna think they're equal.
        # I can't use war.start_time in case prep or something changes
        # and then it will muck up on war attack for a lot of attacks
        return (
            isinstance(other, self.__class__)
            and self.attacker_tag == other.attacker_tag
            and self.defender_tag == other.defender_tag
            and self.order == other.order
        )

    def __init__(self, *, data, client, war):
        self.war = war
        self._raw_data = data if client and client.raw_attribute else None
        self._client = client
        self._from_data(data)

    def _from_data(self, data: dict) -> None:
        self.stars: int = data["stars"]
        self.destruction: float = data["destructionPercentage"]
        self.order: int = data["order"]
        self.attacker_tag: str = data["attackerTag"]
        self.defender_tag: str = data["defenderTag"]
        self.duration: float = data["duration"]

    @property
    def attacker(self) -> "ClanWarMember":
        """:class:`ClanWarMember`: Returns the attacking player."""
        return self.war.get_member(self.attacker_tag)

    @property
    def defender(self) -> "ClanWarMember":
        """:class:`ClanWarMember`: Returns the defending player."""
        return self.war.get_member(self.defender_tag)

    @property
    def is_fresh_attack(self) -> bool:
        """:class:`boolean`: Returns whether the attack is a fresh (first) attack on the defender."""
        if len(self.defender.defenses) == 1:
            # fast route
            return True

        return min(defense.order for defense in self.defender.defenses) == self.order




import inspect
import calendar
import re

from collections import deque, UserDict
from datetime import datetime, timedelta, timezone
from functools import wraps
from operator import attrgetter
from typing import Any, Callable, Generic, Iterable, List, Optional, Type, TypeVar, Union


TAG_VALIDATOR = re.compile(r"^#?[PYLQGRJCUV0289]+$")

T = TypeVar('T')
T_co = TypeVar('T_co', covariant=True)


def find(predicate: Callable[[T], Any], iterable: Iterable[T]) -> Optional[T]:
    """A helper to return the first element found in the sequence
    that meets the predicate.

    For example: ::

        leader = coc.utils.find(lambda m: m.trophies > 5000, clan.members)

    would find the first :class:`~coc.ClanMember` who has more than 5000 trophies and return it.
    If no members have more than 5000 trophies, then ``None`` is returned.

    Parameters
    -----------
    predicate
        A function that returns a boolean-like result.
    iterable: iterable
        The iterable to search through.

    Returns
    -------
    The first item in the iterable which matches the predicate passed.
    """
    for element in iterable:
        if predicate(element):
            return element
    return None


def get(iterable: Iterable[T], **attrs: Any) -> Optional[T]:
    r"""A helper that returns the first item in an iterable that matches the attributes passed.

    If no match is found, ``None`` is returned.

    Example
    -------
    .. code-block:: python3

        member = utils.get(clan.members, level=100, name="Mathsman")
        # returns the first member who has the name "Mathsman" and is level 100

        member = utils.get(clan.members, role=coc.Role.leader)
        # returns the clan leader

        label = utils.get(player.labels, name="Competitive")
        # returns the player's label if they have Competitive.

    Parameters
    ----------
    iterable: iterable
        The list of items to match the attributes from
    \*\*attrs
        A series of kwargs that specify which attributes to match.

    Returns
    -------
    The object from the iterable that matches the attributes passed, or ``None`` if not found.
    """
    converted = [(attrgetter(attr), value) for attr, value in attrs.items()]
    for elem in iterable:
        if all(pred(elem) == value for pred, value in converted):
            return elem
    return None


def from_timestamp(timestamp: str) -> datetime:
    """Parses the raw timestamp given by the API into a :class:`datetime.datetime` object."""
    return datetime.strptime(timestamp, "%Y%m%dT%H%M%S.000Z")


def is_valid_tag(tag: str) -> bool:
    """Validates that a string is a valid Clash of Clans tag.

    This uses the assumption that tags can only consist of the characters PYLQGRJCUV0289.

    Example
    -------

    .. code-block:: python3

        from coc import utils

        user_input = input("Please enter a tag")

        if utils.is_valid_tag(user_input) is True:
            print("{} is a valid tag".format(user_input))
        else:
            print("{} is not a valid tag".format(user_input))

    Returns
    -------
    :class:`bool`
        Whether the tag is a valid tag.
    """
    if TAG_VALIDATOR.match(correct_tag(tag)):
        return True
    return False


def correct_tag(tag: str, prefix: str = "#") -> str:
    """Attempts to correct malformed Clash of Clans tags
    to match how they are formatted in game

    Example
    -------

    .. code-block:: python3

            new_tag = utils.correct_tag(" 123aBc O")
            # new_tag is "#123ABC0".


    Parameters
    ----------
    tag: str
        The tag to correct.
    prefix: str
        The prefix to insert at the start of the tag. Defaults to ``#``.

    Returns
    -------
    str
        The corrected tag.
    """
    return tag and prefix + re.sub(r"[^A-Z0-9]+", "", tag.upper()).replace("O", "0")


def corrected_tag() -> Callable[[Callable[..., T]], Callable[..., T]]:
    """Helper decorator to fix tags passed into client calls. The tag must be the first parameter."""

    def deco(func: Callable[..., T]) -> Callable[..., T]:
        @wraps(func)
        def wrapper(*args, **kwargs) -> T:
            self = args[0]

            if not self.correct_tags:
                return func(*args, **kwargs)

            args = list(args)
            args[1] = correct_tag(args[1])
            return func(*tuple(args), **kwargs)

        return wrapper

    return deco


def maybe_sort(
        seq: Iterable[T], sort: bool, itr: bool = False, key: Callable[[str], Any] = attrgetter("order")
) -> Union[List[T], Iterable[T]]:
    """Returns list or iter based on itr if sort is false otherwise sorted
    with key defaulting to operator.attrgetter('order')
    """
    return (list, iter)[itr](n for n in sorted(seq, key=key)) if sort else (list, iter)[itr](n for n in seq)


def item(
    _object,
    *,
    index: bool = False,
    index_type: Union[int, str] = 0,
    attribute: str = None,
    index_before_attribute: bool = True
):
    """Returns an object, an index, and/or an attribute of the object."""
    attr_get = attrgetter(attribute or "")
    if not (index or index_type or attribute):
        return _object
    if (index or index_type) and not attribute:
        return _object[index_type]
    if attribute and not (index or index_type):
        return attr_get(_object)
    if index_before_attribute:
        return attr_get(_object[index_type])
    return attr_get(_object)[index_type]


def custom_isinstance(obj, module, name):
    """Helper utility to do an `isinstance` check without importing the module (circular imports)"""
    # pylint: disable=broad-except
    for cls in inspect.getmro(type(obj)):
        try:
            if cls.__module__ == module and cls.__name__ == name:
                return True
        except Exception:
            pass
    return False


async def maybe_coroutine(function_, *args, **kwargs):
    """Returns the result of a function which may or may not be a coroutine."""
    value = function_(*args, **kwargs)
    if inspect.isawaitable(value):
        return await value

    return value


def get_season_start(month: Optional[int] = None, year: Optional[int] = None, legacy: bool = False) -> datetime:
    """Get the datetime that the season started.

    This goes by the assumption that SC resets the season on the last monday of every month at 5am UTC.

    .. note::

        If you want the start of the current season, do not pass any parameters in,
        as doing so won't check to ensure the season start is in the past.

    Parameters
    ----------
    month: Optional[int]
        The month to get the season start for. Defaults to the current month/season.
    year: Optional[int]
        The year to get the season start for. Defaults to the current year/season.
    legacy: bool
        Whether to use the old season start method. Defaults to ``False``.

    Returns
    -------
    season_start: :class:`datetime.datetime`
        The start of the season.
    """

    # Start date is the last Monday of the month. That's when SC resets the season values
    def get_old_start_for_month_year(m, y):
        (weekday_of_first_day, days_in_month) = calendar.monthrange(y, m)
        season_start_day = days_in_month - datetime(year=y, month=m, day=days_in_month).weekday()
        return datetime(year=y, month=m, day=season_start_day, hour=5, minute=0, second=0, microsecond=0, tzinfo=timezone.utc)


    # Start date is the first Monday of the month. That's when SC resets the season values
    def get_start_for_month_year(m, y):
        (weekday_of_first_day, days_in_month) = calendar.monthrange(y, m)
        # weekday_of_first_day: 0=Monday, 1=Tuesday, ..., 6=Sunday
        # Calculate days to add to get to the first Monday
        days_until_monday = (7 - weekday_of_first_day) % 7
        season_start_day = 1 + days_until_monday
        return datetime(year=y, month=m, day=season_start_day, hour=5, minute=0, second=0, microsecond=0, tzinfo=timezone.utc)

    start_method = get_start_for_month_year if not legacy else get_old_start_for_month_year

    if month and year:
        # they want a specific month/year combo
        return start_method(month, year)

    now = datetime.now(tz=timezone.utc)
    start = start_method(now.month, now.year)
    if now > start:
        # we got the right one, season started this month
        return start

    # season started last month, so let's try it again.
    if now.month == 1:
        month = 12
        year = now.year - 1
    else:
        month = now.month - 1
        year = now.year
    return start_method(month, year)


def get_season_end(month: Optional[int] = None, year: Optional[int] = None, legacy: bool = False) -> datetime:
    """Get the datetime that the season ends.

    This goes by the assumption that SC resets the season on the last monday of every month at 5am UTC.

    .. note::

        If you want the end of the current season, do not pass any parameters in,
        as doing so won't check to ensure the season end is in the future.

    Parameters
    ----------
    month: Optional[int]
        The month to get the season end for. Defaults to the current month/season.
    year: Optional[int]
        The year to get the season end for. Defaults to the current year/season.
    legacy: bool
        Whether to use the old season end method. Defaults to ``False``.

    Returns
    -------
    season_end: :class:`datetime.datetime`
        The end of the season.
    """
    if month and year:
        if month == 12:
            next_month = 1
            next_year = year + 1
        else:
            next_month = month + 1
            next_year = year
        return get_season_start(next_month, next_year, legacy)

    now = datetime.now(tz=timezone.utc)
    end = get_season_start(now.month, now.year, legacy)
    if end > now:
        return end

    # season ends next month, let's try again
    if now.month == 12:
        month = 1
        year = now.year + 1
    else:
        month = now.month + 1
        year = now.year
    return get_season_start(month, year, legacy)


def get_clan_games_start(time: Optional[datetime] = None) -> datetime:
    """Get the datetime that the next clan games will start.

    This goes by the assumption that clan games start at 8am UTC at the 22nd of each month.

    .. note::

        If you want the start of the next or running clan games, do not pass any parameters in,
        for any other pass a datetime in the month before.

    Parameters
    ----------
    time: Optional[datetime]
        Some time in the month before the clan games you want the start of.

    Returns
    -------
    clan_games_start: :class:`datetime.datetime`
        The start of the next or running clan games.
    """
    if time is None:
        time = datetime.now(tz=timezone.utc).replace(tzinfo=None)
    month = time.month
    year = time.year
    this_months_cg_end = datetime(year=time.year, month=time.month, day=28, hour=8, minute=0, second=0)
    if time > this_months_cg_end and month < 12:
        month += 1
    elif time > this_months_cg_end:  # we're at the end of December
        month = 1
        year += 1
    return datetime(year=year, month=month, day=22, hour=8, minute=0, second=0)


def get_clan_games_end(time: Optional[datetime] = None) -> datetime:
    """Get the datetime that the next clan games will end.

    This goes by the assumption that clan games end at 8am UTC at the 28th of each month.

    .. note::

        If you want the end of the next or running clan games, do not pass any parameters in,
        for any other pass a datetime in the month before.

    Parameters
    ----------
    time: Optional[datetime]
        Some time in the month before the clan games you want the end of.

    Returns
    -------
    clan_games_end: :class:`datetime.datetime`
        The end of the next or running clan games.
    """
    if time is None:
        time = datetime.now(tz=timezone.utc).replace(tzinfo=None)
    month = time.month
    year = time.year
    this_months_cg_end = datetime(year=time.year, month=time.month, day=28, hour=8, minute=0, second=0)
    if time > this_months_cg_end and month < 12:
        month += 1
    elif time > this_months_cg_end:  # we're in December
        month = 1
        year += 1
    return datetime(year=year, month=month, day=28, hour=8, minute=0, second=0)


def get_raid_weekend_start(time: Optional[datetime] = None) -> datetime:
    """Get the datetime that the raid weekend will start.

    This goes by the assumption that raid weekends start at friday 7am UTC.

    .. note::

        If you want the start of the next or running raid weekend, do not pass any parameters in,
        for any other pass a datetime in the week before.

    Parameters
    ----------
    time: Optional[datetime]
        Some time in the week before the raid weekend you want the start of.

    Returns
    -------
    raid_weekend_start: :class:`datetime.datetime`
        The start of the raid weekend.
    """
    if time is None:
        time = datetime.now(tz=timezone.utc).replace(tzinfo=None)
    time = get_raid_weekend_end(time)
    time = time - timedelta(days=3)
    return time


def get_raid_weekend_end(time: Optional[datetime] = None) -> datetime:
    """Get the datetime that the raid weekend will end.

    This goes by the assumption that raid weekends end at monday 7am UTC.

    .. note::

        If you want the end of the next or running raid weekend, do not pass any parameters in,
        for any other pass a datetime in the week before.

    Parameters
    ----------
    time: Optional[datetime]
        Some time in the week before the raid weekend you want the end of.

    Returns
    -------
    raid_weekend_end: :class:`datetime.datetime`
        The end of the raid weekend.
    """
    # Shift the time so that we can pretend the raid ends just after midnight
    if time is None:
        time = datetime.now(tz=timezone.utc).replace(tzinfo=None)
    time = time - timedelta(hours=7, microseconds=1)
    time = time + timedelta(weeks=1, days=-time.weekday())  # Go to the next monday
    time = time.replace(hour=7, minute=0, second=0, microsecond=0)
    return time


class _CachedProperty(Generic[T, T_co]):
    def __init__(self, name: str, function: Callable[[T], T_co]) -> None:
        self.name = name
        self.function = function
        self.__doc__ = getattr(function, '__doc__')

    def __get__(self, instance: T, owner: Type[T]) -> T_co:
        try:
            return getattr(instance, self.name)
        except AttributeError:
            result = self.function(instance)
            setattr(instance, self.name, result)
            return result


def cached_property(name: str) -> Callable[[Callable[[T], T_co]], _CachedProperty[T, T_co]]:
    def deco(func: Callable[[T], T_co]) -> _CachedProperty[T, T_co]:
        return _CachedProperty(name, func)
    return deco


class FIFO(UserDict):
    """Implements a FIFO (least-recently-used) dict with a settable max size."""

    __slots__ = (
        "__keys",
        "max_size",
    )

    def __init__(self, max_size):
        self.max_size = max_size
        self.__keys = deque()
        super().__init__()

    def __verify_max_size(self):
        while len(self) > self.max_size:
            del self[self.__keys.popleft()]

    def __setitem__(self, key, value):
        self.__keys.append(key)
        super().__setitem__(key, value)
        self.__verify_max_size()

    def __getitem__(self, key):
        self.__verify_max_size()
        return super().__getitem__(key)

    def __contains__(self, key):
        self.__verify_max_size()
        return super().__contains__(key)

    def copy(self):
        self.data = self.data.copy()
        return self


class HTTPStats(dict):
    """Implements a basic key: deque value to aid with HTTP performance stats."""

    __slots__ = ("max_size",)

    def __init__(self, max_size):
        self.max_size = max_size
        super().__init__()

    def __setitem__(self, key, value):
        try:
            super().__getitem__(key).append(value)
        except (KeyError, AttributeError):
            super().__setitem__(key, deque((value,), maxlen=self.max_size))

    def get_average(self, key):
        """Get the average latency / performance counter for an API endpoint"""
        try:
            stats = self[key]
        except KeyError:
            return None

        return sum(stats) / len(stats)

    def get_mixed_average(self):
        """Get the average latency / performance counter for all API endpoints"""
        stats = [*self.values()]
        return sum(stats) / len(stats)

    def get_all_average(self):
        """Get the average latency / performance counter for each API endpoint."""
        return {k: sum(v) / len(v) for k, v in self.items()}


class CaseInsensitiveDict(dict):
    def __getitem__(self, key):
        if isinstance(key, tuple):
            key = tuple(x.lower() if isinstance(x, str) else x for x in key)
        else:
            key = key.lower()

        return super().__getitem__(key)

    def get(self, key, default=None):
        if isinstance(key, tuple):
            key = tuple(x.lower() if isinstance(x, str) else x for x in key)
        else:
            key = key.lower()

        return super().get(key, default)

    def __setitem__(self, key, v):
        if isinstance(key, tuple):
            key = tuple(x.lower() if isinstance(x, str) else x for x in key)
        else:
            key = key.lower()

        super().__setitem__(key, v)







def _get_maybe_first(dict_items, lookup, default=None):
    try:
        items = dict_items[lookup]
    except KeyError:
        return default
    else:
        try:
            return items[0]
        except (IndexError, KeyError):
            return default




import typing


from .players import ClanMember
from .miscmodels import try_enum, ChatLanguage, Location, Label, BaseLeague, CapitalDistrict
from .utils import get, cached_property, correct_tag
from .enums import ClanType
from .abc import BaseClan


class RankedClan(BaseClan):
    """Represents the clan object returned by leader-board rankings.

    Attributes
    ----------
    tag: :class:`str`
        The clan's tag
    name: :class:`str`
        The clan's name
    badge: :class:`Badge`
        The clan's badge
    level: :class:`int`
        The clan's level.
    location: :class:`Location`
        The clan's location.
    member_count: :class:`int`
        The number of members in the clan.
    points: :class:`int`
        The clan's trophy-count. If retrieving info for capital or builder base leader-boards, this will be ``None``.
    builder_base_points: :class:`int`
        The clan's builder base trophy count. If retrieving info for regular or capital leader boards, this will be
        ``None``.
    capital_points: :class:`int`
        The clan's capital trophy count. If retrieving info for regular or builder base leader boards, this will be
        ``None``.
    rank: :class:`int`
        The clan's rank in the leader board.
    previous_rank: :class:`int`
        The clan's rank in the previous season's leaderboard.
    """

    __slots__ = (
        "location",
        "member_count",
        "points",
        "builder_base_points",
        "capital_points",
        "rank",
        "previous_rank",
    )

    def __init__(self, *, data, client, **_):
        super().__init__(data=data, client=client)
        self._from_data(data)

    def _from_data(self, data: dict) -> None:
        data_get = data.get

        self.points: int = data_get("clanPoints")
        self.builder_base_points: int = data_get("clanBuilderBasePoints")
        self.capital_points: int = data_get("clanCapitalPoints")
        self.member_count: int = data_get("members")
        self.location = try_enum(Location, data=data_get("location"))
        self.rank: int = data_get("rank")
        self.previous_rank: int = data_get("previousRank")


class Clan(BaseClan):
    """
    Represents a clan in Clash of Clans.

    Attributes
    ----------
    tag: :class:`str`
        The clan's tag
    name: :class:`str`
        The clan's name
    badge: :class:`Badge`
        The clan's badge
    level: :class:`int`
        The clan's level.
    type: :class:`ClanType`
        The clan's type for accepting members.
        This could be ``open``, ``inviteOnly`` or ``closed``.
    family_friendly: :class:`bool`
        Indicates the clan's family friendly status. This can be True if family friendly, or False if not.
    description: :class:`str`
        The clan's public description.
    location: :class:`Location`
        The clan's location.
    points: :class:`int`
        The clan's trophy count. This is calculated according to members' trophy counts.
    builder_base_points: :class:`int`
        The clan's builder base trophy count. This is calculated according to members' builder base trophy counts.
    capital_points: :class:`int`
        The clan's clan capital points. Unsure how this is calculated.
    required_trophies: :class:`int`
        The minimum trophies required to apply to this clan.
    required_builder_base_trophies: :class:`int`
        The minimum builder base trophies required to apply to this clan.
    required_townhall: :class:`int`
        The minimum townhall level required to apply to this clan.
    war_frequency: :class:`str`
        The frequency for when this clan goes to war.
        For example, this could be ``always``.
    war_win_streak: :class:`int`
        The clan's current war winning streak.
    war_wins: :class:`int`
        The number of wars the clan has won.
    war_ties: :class:`int`
        The number of wars the clan has tied. This is only available from the clan search endpoint else -1.
    war_losses: :class:`int`
        The number of wars the clan has lost.  This is only available from the clan search endpoint else -1.
    public_war_log: :class:`bool`
        Indicates if the clan has a public war log.
        If this is ``False``, operations to find the clan's current
        war may raise :exc:`PrivateWarLog`.
    member_count: :class:`int`
        The number of members in the clan.
    label_cls: :class:`coc.Label`
        The type which the labels found in :attr:`Clan.labels` will be of.
        Ensure any overriding of this inherits from :class:`coc.Label`.
    member_cls: :class:`coc.ClanMember`
        The type which the members found in :attr:`Clan.members` will be of.
        Ensure any overriding of this inherits from :class:`coc.ClanMember`.
    capital_district_cls: :class:`coc.CapitalDistrict`
        The type which the clan capital districts found in
        :attr:`Clan.capital_districts` will be of. Ensure any overriding of
        this inherits from :class:`coc.CapitalDistrict`.
    war_league: :class:`coc.BaseLeague`
        The clan's CWL league.
    capital_league: :class:`coc.BaseLeague`
        The clan's Clan Capital league.
    """

    __slots__ = (
        "type",
        "family_friendly",
        "description",
        "location",
        "points",
        "builder_base_points",
        "capital_points",
        "required_trophies",
        "required_builder_base_trophies",
        "war_frequency",
        "war_win_streak",
        "war_wins",
        "war_ties",
        "war_losses",
        "public_war_log",
        "member_count",
        "_labels",
        "_members",
        "_districts",
        "_client",
        "label_cls",
        "member_cls",
        "capital_district_cls",
        "war_league",
        "capital_league",
        "chat_language",
        "required_townhall",

        "_cs_labels",
        "_cs_members",
        "_cs_members_dict",
        "_cs_capital_districts",
        "_iter_labels",
        "_iter_members",
        "_iter_capital_districts"
    )

    def __init__(self, *, data, client, **_):
        super().__init__(data=data, client=client)
        self.label_cls = Label
        self.member_cls = ClanMember
        self.capital_district_cls = CapitalDistrict

        self._from_data(data)

    def _from_data(self, data: dict) -> None:
        data_get = data.get

        self.points: int = data_get("clanPoints")
        self.builder_base_points: int = data_get("clanBuilderBasePoints")
        self.capital_points: int = data_get("clanCapitalPoints")
        self.member_count: int = data_get("members")
        self.location = try_enum(Location, data=data_get("location"))

        # only available via /clans/{clanTag} or /clans endpoint
        self.type = ClanType(value=data_get("type"))
        self.family_friendly = data_get("isFamilyFriendly")
        self.required_trophies: int = data_get("requiredTrophies")
        self.required_builder_base_trophies: int = data_get("requiredBuilderBaseTrophies")
        self.war_frequency: str = data_get("warFrequency")
        self.war_win_streak: int = data_get("warWinStreak")
        self.war_wins: int = data_get("warWins")
        self.war_ties: int = data_get("warTies", -1)
        self.war_losses: int = data_get("warLosses", -1)
        self.public_war_log: bool = data_get("isWarLogPublic")
        self.description: str = data_get("description")
        self.war_league = try_enum(BaseLeague, data=data_get("warLeague"))
        self.capital_league = try_enum(BaseLeague, data=data_get("capitalLeague"))
        self.chat_language = try_enum(ChatLanguage, data=data_get("chatLanguage"))
        self.required_townhall = data_get("requiredTownhallLevel")

        label_cls = self.label_cls
        self._iter_labels = (label_cls(data=ldata, client=self._client) for ldata in data_get("labels", []))

        # update members globally. only available via /clans/{clanTag}
        member_cls = self.member_cls
        member_data = data.get("memberList", [])
        for rank, mdata in enumerate(sorted(member_data, key=lambda x: x["builderBaseTrophies"], reverse=True), 1):
            mdata["builderBaseRank"] = rank
        self._iter_members = (
            member_cls(data=mdata, client=self._client, clan=self) for mdata in member_data
        )

        capital_district_cls = self.capital_district_cls
        if data_get("clanCapital"):
            self._iter_capital_districts = (capital_district_cls(data=cddata, client=self._client) for cddata in
                                            data_get("clanCapital")["districts"])
        else:
            self._iter_capital_districts = ()

    @cached_property("_cs_labels")
    def labels(self) -> typing.List[Label]:
        """List[:class:`Label`]: A list of :class:`Label` objects that the clan has."""
        return list(self._iter_labels)

    @cached_property("_cs_members")
    def members(self) -> typing.List[ClanMember]:
        """List[:class:`ClanMember`]: A list of members that belong to the clan."""
        return list(self.members_dict.values())

    @cached_property("_cs_members_dict")
    def members_dict(self) -> typing.Dict[str, ClanMember]:
        """Dict[str, :class:`ClanMember`]: A dict of members that belong to the clan."""
        return {m.tag: m for m in self._iter_members}


    @cached_property("_cs_capital_districts")
    def capital_districts(self) -> typing.List[CapitalDistrict]:
        """List[:class:`CapitalDistrict`]: A :class:`List` of :class:`CapitalDistrict` that the clan has."""
        return list(self._iter_capital_districts)

    def get_member(self, tag: str) -> typing.Optional[ClanMember]:
        """Return a :class:`ClanMember` with the tag provided. Returns ``None`` if not found.

        Example
        --------
        .. code-block:: python3

            clan = await client.get_clan('clan_tag')
            member = clan.get_member('player_tag')

        Returns
        --------
        The member who matches the tag provided: Optional[:class:`ClanMember`]
        """
        tag = correct_tag(tag)
        if not self.members_dict:
            _ = self.members

        try:
            return self.members_dict[tag]
        except KeyError:
            return None

    def get_member_by(self, **attrs) -> typing.Optional[ClanMember]:
        """Returns the first :class:`ClanMember` that meets the attributes passed

        This search implements the :func:`coc.utils.get` function

        Example
        -------
        .. code-block:: python3

            clan = await client.get_clan("#clantag")
            member = clan.get_member_by(name="Joe")
            member = clan.get_member_by(level=125)
            member = clan.get_member_by(role=coc.Role.elder)

        """
        return get(self.members, **attrs)




from .abc import BaseDataClass
from .enums import Resource, PlayerHouseElementType, VillageType, SceneryType
from .miscmodels import TID


class Decoration(BaseDataClass):
    """Represents a Decoration.

    Attributes
    ----------
    id: :class:`int`
        The decoration's unique identifier.
    name: :class:`str`
        The decoration's name.
    TID: :class:`TID`
        The decoration's translation IDs for localization.
    width: :class:`int`
        The width of the decoration.
    in_shop: :class:`bool`
        Whether this decoration is available in the shop.
    pass_reward: :class:`bool`
        Whether this decoration is a pass reward.
    max_count: :class:`int`
        The maximum number of this decoration that can be placed.
    shop_resource: :class:`Resource`
        The resource type required to buy this decoration.
    cost: :class:`int`
        The cost to buy/place this decoration.
    village: :class:`VillageType`
        The village type where this decoration belongs.
    """

    __slots__ = (
        "id",
        "name",
        "TID",
        "width",
        "in_shop",
        "pass_reward",
        "max_count",
        "shop_resource",
        "cost",
        "village",
    )

    def __init__(self, data: dict):
        self.id: int = data["_id"]
        self.name: str = data["name"]
        self.TID: TID = TID(data=data["TID"])
        self.width: int = data["width"]
        self.in_shop: bool = not data["not_in_shop"]
        self.pass_reward: bool = data["pass_reward"]
        self.max_count: int = data["max_count"]
        self.shop_resource: Resource = Resource(value=data["build_resource"])
        self.cost: int = data["build_cost"]
        self.village = VillageType(value=data["village"])

class Obstacle(BaseDataClass):
    """Represents an Obstacle.

    Attributes
    ----------
    id: :class:`int`
        The obstacle's unique identifier.
    name: :class:`str`
        The obstacle's name.
    TID: :class:`TID`
        The obstacle's translation IDs for localization.
    width: :class:`int`
        The width of the obstacle.
    clear_resource: :class:`Resource`
        The resource type required to clear this obstacle.
    clear_cost: :class:`int`
        The cost to clear this obstacle.
    loot_resource: Optional[:class:`Resource`]
        The resource type obtained from clearing this obstacle.
    loot_count: Optional[:class:`int`]
        The amount of loot obtained from clearing this obstacle.
    village: :class:`VillageType`
        The village type where this obstacle belongs.
    """

    __slots__ = (
        "id",
        "name",
        "TID",
        "width",
        "clear_resource",
        "clear_cost",
        "loot_resource",
        "loot_count",
        "village",
    )

    def __init__(self, data: dict):
        self.id: int = data["_id"]
        self.name: str = data["name"]
        self.TID: TID = TID(data=data["TID"])
        self.width: int = data["width"]
        self.clear_resource = Resource(value=data["clear_resource"])
        self.clear_cost: int = data["clear_cost"]
        self.loot_resource : Resource | None = data["loot_resource"] and Resource(value=data["loot_resource"])
        self.loot_count: int | None = data["loot_count"]
        self.village = VillageType(value=data["village"])

class Scenery(BaseDataClass):
    """Represents a Scenery.

    Attributes
    ----------
    id: :class:`int`
        The scenery's unique identifier.
    name: :class:`str`
        The scenery's name.
    TID: :class:`TID`
        The scenery's translation IDs for localization.
    village: :class:`SceneryType`
        The scenery type.
    has_music: :class:`bool`
        Whether this scenery has custom music.
    """

    __slots__ = (
        "id",
        "name",
        "TID",
        "village",
        "has_music",
    )

    def __init__(self, data: dict):
        self.id: int = data["_id"]
        self.name: str = data["name"]
        self.TID: TID = TID(data=data["TID"])
        self.village = SceneryType(value=data["type"])
        self.has_music: bool = data["music"] is not None

class Skin(BaseDataClass):
    """Represents a Hero Skin.

    Attributes
    ----------
    id: :class:`int`
        The skin's unique identifier.
    name: :class:`str`
        The skin's name.
    TID: :class:`TID`
        The skin's translation IDs for localization.
    tier: :class:`str`
        The tier/rarity of the skin.
    hero: :class:`str`
        The hero this skin belongs to.
    """

    __slots__ = (
        "id",
        "name",
        "TID",
        "tier",
        "hero",
    )

    def __init__(self, data: dict):
        self.id: int = data["_id"]
        self.name: str = data["name"]
        self.TID: TID = TID(data=data["TID"])
        self.tier: str = data["tier"]
        self.hero: str = data["character"]

class ClanCapitalHousePart(BaseDataClass):
    """Represents a Clan Capital House Part.

    Attributes
    ----------
    id: :class:`int`
        The house part's unique identifier.
    name: :class:`str`
        The house part's name.
    slot_type: :class:`PlayerHouseElementType`
        The type of house element slot.
    pass_reward: :class:`bool`
        Whether this house part is a pass reward.
    """

    __slots__ = (
        "id",
        "name",
        "slot_type",
        "pass_reward",
    )

    def __init__(self, data: dict):
        self.id: int = data["_id"]
        self.name: str = data["name"]
        self.slot_type = PlayerHouseElementType(value=data["slot_type"])
        self.pass_reward: bool = data["pass_reward"]





import typing

from .abc import BaseClan
from .war_members import ClanWarMember, ClanWarLeagueClanMember
from .utils import cached_property, correct_tag

if typing.TYPE_CHECKING:
    from .war_attack import WarAttack  # noqa


class WarClan(BaseClan):
    """Represents a War Clan that the API returns.

    .. note::

        If this is called via :attr:`ClanWarLog.clan`, then :attr:`WarClan.members`,
        :attr:`WarClan.attacks` and :attr:`WarClan.defenses` will be empty.

        If this is called via :attr:`ClanWarLog.opponent`, then :attr:`WarClan.members`,
        :attr:`WarClan.attacks` and :attr:`WarClan.defenses` will be empty.
        :attr:`WarClan.exp_earned` and :attr:`WarClan.attacks` will be ``None``.

        If this is called via :attr:`ClanWar.clan` or :attr:`ClanWar.opponent` then
        :attr:`WarClan.exp_earned` will be ``None``.

    Attributes
    ----------
    tag: :class:`str`
        The clan's tag
    name: :class:`str`
        The clan's name
    badge: :class:`Badge`
        The clan's badge
    level: :class:`int`
        The clan's level.
    stars:
        :class:`int`: Number of stars by clan this war.
    destruction:
        :class:`float`: Destruction as a percentage.
    exp_earned:
        :class:`int`: Total XP earned by clan this war.
    attacks_used:
        :class:`int`: Total attacks used by clan this war.
    max_stars:
        :class:`int`: Total possible stars achievable.
    total_attacks:
        :class:`int`: Total possible number of attacks usable by clan this war
    member_cls: :class:`coc.ClanMember`
        The type which the members found in :attr:`Clan.members` will be of.
        Ensure any overriding of this inherits from :class:`coc.ClanMember`.
    """

    __slots__ = (
        "destruction",
        "level",
        "attacks_used",
        "stars",
        "exp_earned",
        "max_stars",
        "total_attacks",
        "member_cls",
        "_war",
        "_client",

        "_members",
        "_iter_members",
        "_cs_attacks",
        "_cs_defenses",
        "_cs_members",
    )

    def __init__(self, *, data, client, war, **kwargs):
        self._war = war
        self._members = {}
        self.member_cls = kwargs.pop('member_cls', ClanWarMember)

        super().__init__(data=data, client=client)
        self._from_data(data)

    def _from_data(self, data: dict) -> None:
        data_get = data.get

        self.level: int = data_get("clanLevel")
        self.destruction: float = data_get("destructionPercentage")
        self.exp_earned: int = data_get("expEarned")
        self.attacks_used: int = data_get("attacks")
        self.stars: int = data_get("stars")

        if self._war:
            self.max_stars: int = self._war.team_size * 3
            self.total_attacks: int = self._war.team_size * self._war.attacks_per_member
        else:
            self.max_stars: int = (data_get("teamSize") or 0) * 3
            self.total_attacks: int = (data_get("teamSize") or 0) * 3

        self._iter_members = (
            self.member_cls(data=mdata, client=self._client, war=self._war, clan=self)
            for mdata in data_get("members", [])
        )

    @cached_property("_cs_members")
    def members(self) -> typing.List[ClanWarMember]:
        """List[:class:`ClanWarMember`]: A list of members that are in the war.
        This is sorted by :attr:`ClanWarMember.map_position`
        """
        dict_members = self._members = {m.tag: m for m in sorted(self._iter_members, key=lambda m: m.map_position)}
        return list(dict_members.values())

    @property
    def is_opponent(self) -> bool:
        """:class:`bool`: Indicates whether the clan is the opponent."""
        return self.tag == self._war.opponent.tag

    @cached_property("_cs_attacks")
    def attacks(self) -> typing.List["WarAttack"]:
        """List[:class:`WarAttack`]: Returns all clan member's attacks this
        war. This is sorted by attack order."""
        if not self._war:
            return []

        attacks = []
        for member in self.members:
            attacks.extend(member.attacks)

        return list(sorted(attacks, key=lambda a: a.order, reverse=True))

    @cached_property("_cs_defenses")
    def defenses(self) -> typing.List["WarAttack"]:
        """List[:class:`WarAttack`]: Returns all clan member's defenses
        this war. This is sorted by attack order.

        Equivalent to the other team's ``.attacks`` property.
        """
        other = self._war and self._war.clan if self.is_opponent else self._war.opponent
        return other.attacks if other else []
    
    @property
    def average_attack_duration(self) -> int:
        """:class:`int`: Returns the average duration of all clan member's
        attacks this war."""
        count = len(self.attacks)
        total_duration = sum(attack.duration for attack in self.attacks)
        return int(total_duration/count) if count > 0 else 0

    def get_member(self, tag: str) -> typing.Optional[ClanWarMember]:
        """
        Get a member of the clan for the given tag, or ``None`` if not found.

        Returns
        --------
        ClanWarMember or None
            The clan member who matches the tag.: Optional[:class:`ClanWarMember`]"""

        tag = correct_tag(tag)
        if not self._members:
            _ = self.members

        try:
            return self._members[tag]
        except KeyError:
            return None


class ClanWarLeagueClan(BaseClan):
    """Represents a Clan War League Clan.

    Attributes
    ----------
    tag: :class:`str`
        The clan's tag
    name: :class:`str`
        The clan's name
    badge: :class:`Badge`
        The clan's badge
    level: :class:`int`
        The clan's level.
    """

    __slots__ = BaseClan.__slots__ + ("_cs_members", "_iter_members")

    def __init__(self, *, data, client):
        super().__init__(data=data, client=client)

        self._iter_members = (
            ClanWarLeagueClanMember(data=mdata, client=self._client) for mdata in data.get("members", [])
        )

    @cached_property("_cs_members")
    def members(self) -> typing.List[ClanWarLeagueClanMember]:
        """List[:class:`ClanWarLeagueClanMember`]: A list of players participating in this clan war league season.

        This list is selected when the clan chooses to participate in CWL, and will not change throughout the season.
        It is sometimes referred to as the `master roster`.
        """
        return list(self._iter_members)




from pathlib import Path
from typing import AsyncIterator, Type, TYPE_CHECKING

from .miscmodels import try_enum, Badge
from .iterators import PlayerIterator

if TYPE_CHECKING:
    from .players import Player


class BaseClan:
    """
    Abstract data class that represents base Clan objects

    Attributes
    ----------
    tag: :class:`str`
        The clan's tag
    name: :class:`str`
        The clan's name
    badge: :class:`Badge`
        The clan's badge
    level: :class:`int`
        The clan's level.
    """
    __slots__ = ("tag", "name", "_client", "badge", "level", "_response_retry", "_raw_data")

    def __init__(self, *, data, client, **kwargs):
        self._client = client

        self._response_retry = data.get("_response_retry")
        self.tag = data.get("tag")
        self.name = data.get("name")
        self.badge = try_enum(Badge, data=data.get("badgeUrls"),
                              client=self._client)
        self.level = data.get("clanLevel")
        self._raw_data = data if client and client.raw_attribute else None

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"<{self.__class__.__name__} tag={self.tag} name={self.name}>"

    def __eq__(self, other):
        return isinstance(other, BaseClan) and self.tag == other.tag

    @property
    def share_link(self) -> str:
        """str: A formatted link to open the clan in-game"""
        return f"https://link.clashofclans.com/en?action=OpenClanProfile&tag=%23{self.tag.strip('#')}"

    @property
    def members(self):
        # pylint: disable=missing-function-docstring
        return NotImplemented

    def get_detailed_members(self, cls: Type["Player"] = None,
                             load_game_data: bool = None) -> AsyncIterator["Player"]:
        """Get detailed player information for every player in the clan.

        This returns a :class:`PlayerIterator` which fetches all player
        tags in the clan in parallel.

        Example
        ---------

        .. code-block:: python3

            clan = await client.get_clan(clan_tag)

            async for player in clan.get_detailed_members():
                print(player.name)


        Yields
        ------
        :class:`Player`
            A full player object of a clan member.
        """
        if self.members is NotImplemented:
            return NotImplemented
        if load_game_data and not isinstance(load_game_data, bool):
            raise TypeError("load_game_data must be either True or False.")

        return PlayerIterator(self._client,
                              (p.tag for p in self.members),
                              cls=cls,
                              load_game_data=load_game_data,
                              members=self.members_dict)


class BasePlayer:
    """An ABC that implements some common operations on players, regardless of type.

    Attributes
    ----------
    tag: :class:`str`
        The player's tag
    name: :class:`str`
        The player's name
    """

    __slots__ = ("tag", "name", "_client", "_response_retry", "_raw_data")

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"<{self.__class__.__name__} tag={self.tag} name={self.name}>"

    def __eq__(self, other):
        return isinstance(other, BasePlayer) and self.tag == other.tag

    def __init__(self, *, data, client, **_):
        self._client = client
        self._response_retry = data.get("_response_retry")
        self._raw_data = data if client and client.raw_attribute else None
        self.tag = data.get("tag")
        self.name = data.get("name")

    @property
    def share_link(self) -> str:
        """:class:`str` - A formatted link to open the player in-game"""
        return f"https://link.clashofclans.com/en?action=OpenPlayerProfile&tag=%23{self.tag.strip('#')}"


class BaseDataClass:
    """Base class for simple data classes with standard repr.
    
    This class provides a default __repr__ implementation that displays
    the name and id attributes if they exist.
    """

    __slots__ = ()

    def __repr__(self):
        attrs = [
            ("name", getattr(self, "name", "Unknown")),
            ("id", getattr(self, "id", None)),
        ]
        # Filter out None values
        attrs = [(k, v) for k, v in attrs if v is not None]
        return "<%s %s>" % (
            self.__class__.__name__, " ".join("%s=%r" % t for t in attrs))


class LevelManager:
    """Handles level property and level data loading.
    
    This class manages the level state and triggers level-specific data loading
    when the level changes. Subclasses must implement _load_level_data().
    
    Attributes
    ----------
    level: :class:`int`
        The current level of the unit.
    max_level: :class:`int`
        The maximum level this unit can reach.
    is_max: :class:`bool`
        Whether the unit is at its maximum level.
    """

    __slots__ = (
        "_level",
        "_static_data",
        "max_level",
    )

    def __init__(self, initial_level: int, static_data: dict | None):

        #if the level is set to 0, initalize it as the lowest available level
        if initial_level == 0 and static_data:
            min_level = 1
            if static_data.get("levels"):
                min_level = static_data["levels"][0].get("level")
            initial_level = min_level

        self._level: int = initial_level
        self._static_data = static_data

        self.max_level: int = len(static_data["levels"]) if static_data else 0

    @property
    def level(self) -> int:
        """:class:`int`: The current level of the unit."""
        return self._level

    @level.setter
    def level(self, value: int):
        if not self._static_data:
            self._level = value
            return

        if value < 1:
            raise ValueError(f"Level must be greater than 1")

        self._level = value
        self._load_level_data()

    @property
    def is_max(self) -> bool:
        """:class:`bool`: Whether the unit is at its maximum level."""
        return self.max_level == self.level

    def _load_level_data(self):
        raise NotImplementedError("Subclasses must implement _load_level_data()")

    def __getattribute__(self, name):
        try:
            return object.__getattribute__(self, name)
        except AttributeError:
            if name == "_raw_data":
                raise AttributeError(
                    f"'{self.__class__.__name__}' object has no attribute '{name}'. "
                    f"This attribute requires client raw_attribute to be enabled and set to True."
                    f"Additionally must be an API data filled unit."
                )
            else:
                raise AttributeError(
                    f"'{self.__class__.__name__}' object has no attribute '{name}'. "
                    f"This attribute requires static game data to be loaded. "
                    f"Use load_game_data=True when fetching this object."
                )


    def __repr__(self):
        attrs = [
            ("name", getattr(self, "name", "Unknown")),
            ("level", self.level),
        ]
        if hasattr(self, "id"):
            attrs.insert(1, ("id", self.id))
        return "<%s %s>" % (
            self.__class__.__name__, " ".join("%s=%r" % t for t in attrs))


class LeveledUnit(LevelManager):
    """Base class for units with level-dependent data and additional utilities.
    
    Extends LevelManager with townhall-specific level queries.
    """

    __slots__ = ()

    def get_max_level_for_townhall(self, townhall: int) -> int | None:
        """Get the maximum level for this unit at a given townhall level.

        Parameters
        ----------
        townhall: :class:`int`
            The townhall level to check.

        Returns
        -------
        Optional[:class:`int`]
            The maximum level this unit can reach at the given townhall level,
            or ``None`` if no static data is available.
        """
        if not self._static_data or not self._static_data["levels"]:
            return None

        max_level = None
        for level_data in self._static_data.get("levels", []):
            if level_data.get("required_townhall", 0) <= townhall:
                max_level = level_data.get("level")
            else:
                break

        return max_level